#include "byteorder.h"
#include "Exceptions.h"
#include "BinaryAdapterFormat.h"

using namespace phylum;

POOL_DEFINE(BinaryAdapterFormat, 1024)

// Add a single element
void BinaryAdapterFormat::addField(const std::string& name, ElementType type)
{
  nameTable_.insert(std::make_pair(name, fields_.size()));
  fields_.push_back(Field(ELEMENT, type));
}

// Add an array of elements
void BinaryAdapterFormat::addArrayField(const std::string& name, ElementType type, size_t length)
{
  nameTable_.insert(std::make_pair(name, fields_.size()));
  fields_.push_back(Field(ARRAY, type, length));
}

// Add a variable length array of elements
void BinaryAdapterFormat::addVarArrayField(const std::string& name, ElementType type, const std::string& elementName)
{
  nameTable_.insert(std::make_pair(name, fields_.size()));
  fields_.push_back(Field(VARARRAY, type, elementName));
}

// Add a terminated variable length array of elements
void BinaryAdapterFormat::addTermArrayField(const std::string& name, ElementType type, double terminator)
{
  nameTable_.insert(std::make_pair(name, fields_.size()));
  fields_.push_back(Field(TERMARRAY, type, terminator));
}

// Add a fixed length string
void BinaryAdapterFormat::addStringField(const std::string& name, size_t length)
{
  nameTable_.insert(std::make_pair(name, fields_.size()));
  fields_.push_back(Field(STRING, INT8, length));
}

// Add a variable length string
void BinaryAdapterFormat::addVarStringField(const std::string& name, const std::string& elementName)
{
  nameTable_.insert(std::make_pair(name, fields_.size()));
  fields_.push_back(Field(VARSTRING, INT8, elementName));
}

// Add a terminates variable length string
void BinaryAdapterFormat::addTermStringField(const std::string& name, double terminator)
{
  nameTable_.insert(std::make_pair(name, fields_.size()));
  fields_.push_back(Field(TERMSTRING, INT8, terminator));
}

// Add a padding field of one or more bytes
void BinaryAdapterFormat::addPaddingField(size_t size)
{
  fields_.push_back(Field(ARRAY, INT8, size));
}

// Add a named padding field; experience shows that some times values are hidden in padding
void BinaryAdapterFormat::addPaddingField(const std::string& name, size_t size)
{
  nameTable_.insert(std::make_pair(name, fields_.size()));
  fields_.push_back(Field(ARRAY, INT8, size));
}

size_t BinaryAdapterFormat::sizeOf(ElementType type) const
{
  switch(type)
  {
  case INT8:
  case UINT8:
    return 1;
  case INT16:
  case UINT16:
    return 2;
  case INT32:
  case UINT32:
  case FLOAT32:
    return 4;
  case INT64:
  case UINT64:
  case FLOAT64:
    return 8;
  default:
    throw InvalidTypeException("Unrecognized element type");
  }
}

size_t BinaryAdapterFormat::getVarLength(size_t pos) const
{
  if(fields_[pos].fieldType_ != ELEMENT)
  {
    throw InvalidTypeException("The field containing the variable length value must be of type ELEMENT");
  }

  switch(fields_[pos].elementType_)
  {
  case INT8:
    return getLength<int8_t>(pos);
  case UINT8:
    return getLength<uint8_t>(pos);
  case INT16:
    return getLength<int16_t>(pos);
  case UINT16:
    return getLength<uint16_t>(pos);
  case INT32:
    return getLength<int32_t>(pos);
  case UINT32:
    return getLength<uint32_t>(pos);
  case FLOAT32:
    return getLength<float>(pos);
  case INT64:
    return getLength<int64_t>(pos);
  case UINT64:
    return getLength<uint64_t>(pos);
  case FLOAT64:
    return getLength<double>(pos);
  default:
    throw InvalidTypeException("Unrecognized element type");
  }
}

void BinaryAdapterFormat::setVarLength(size_t pos, size_t length)
{
  if(fields_[pos].fieldType_ != ELEMENT)
  {
    throw InvalidTypeException("The field containing the variable length value must be of type ELEMENT");
  }

  switch(fields_[pos].elementType_)
  {
  case INT8:
    setFieldValue(pos, static_cast<int8_t>(length));
    break;
  case UINT8:
    setFieldValue(pos, static_cast<uint8_t>(length));
    break;
  case INT16:
    setFieldValue(pos, static_cast<int16_t>(length));
    break;
  case UINT16:
    setFieldValue(pos, static_cast<uint16_t>(length));
    break;
  case INT32:
    setFieldValue(pos, static_cast<int32_t>(length));
    break;
  case UINT32:
    setFieldValue(pos, static_cast<uint32_t>(length));
    break;
  case FLOAT32:
    setFieldValue(pos, static_cast<float>(length));
    break;
  case INT64:
    setFieldValue(pos, static_cast<int64_t>(length));
    break;
  case UINT64:
    setFieldValue(pos, static_cast<uint64_t>(length));
    break;
  case FLOAT64:
    setFieldValue(pos, static_cast<double>(length));
    break;
  default:
    throw InvalidTypeException("Unrecognized element type");
  }
}

size_t BinaryAdapterFormat::advanceToTerminator(BinaryAdapterFormat::ElementType type, const char* buffer, size_t size, double sentinal) const
{
  switch(type)
  {
  case INT8:
    return advanceTo<int8_t>(buffer, size, sentinal);
  case UINT8:
    return advanceTo<uint8_t>(buffer, size, sentinal);
  case INT16:
    return advanceTo<int16_t>(buffer, size, sentinal);
  case UINT16:
    return advanceTo<uint16_t>(buffer, size, sentinal);
  case INT32:
    return advanceTo<int32_t>(buffer, size, sentinal);
  case UINT32:
    return advanceTo<uint32_t>(buffer, size, sentinal);
  case FLOAT32:
    return advanceTo<float>(buffer, size, sentinal);
  case INT64:
    return advanceTo<int64_t>(buffer, size, sentinal);
  case UINT64:
    return advanceTo<uint64_t>(buffer, size, sentinal);
  case FLOAT64:
    return advanceTo<double>(buffer, size, sentinal);
  default:
    throw InvalidTypeException("Unrecognized element type");
  }
}

void BinaryAdapterFormat::appendTerminator(BinaryAdapterFormat::ElementType type, char* buffer, double sentinal) const
{
  // Convert the sentinal to the correct type and append to buffer
  switch(type)
  {
  case INT8:
    appendTo<int8_t>(buffer, sentinal);
    break;
  case UINT8:
    appendTo<uint8_t>(buffer, sentinal);
    break;
  case INT16:
    appendTo<int16_t>(buffer, sentinal);
    break;
  case UINT16:
    appendTo<uint16_t>(buffer, sentinal);
    break;
  case INT32:
    appendTo<int32_t>(buffer, sentinal);
    break;
  case UINT32:
    appendTo<uint32_t>(buffer, sentinal);
    break;
  case FLOAT32:
    appendTo<float>(buffer, sentinal);
    break;
  case INT64:
    appendTo<int64_t>(buffer, sentinal);
    break;
  case UINT64:
    appendTo<uint64_t>(buffer, sentinal);
    break;
  case FLOAT64:
    appendTo<double>(buffer, sentinal);
    break;
  default:
    throw InvalidTypeException("Unrecognized element type");
  }
}

// Adapt buffer to format
void BinaryAdapterFormat::adapt(const char* buffer, size_t size)
{
  size_t start, fieldSize;
  size_t index = 0;

  // Copy the buffer contents to the fields
  for(Fields::iterator iter = fields_.begin(); iter != fields_.end(); ++iter)
  {
    // Determine the start position and length of the current field
    switch((*iter).fieldType_)
    {
    case ELEMENT:
    {
      start = index;
      fieldSize = sizeOf((*iter).elementType_);
      index += fieldSize;
      if(size < index)
      {
        throw SizeException("Insufficient data for adaptaion");
      }

      (*iter).field_.assign(&buffer[start], &buffer[index]);
      break;
    }
    case ARRAY:
    case STRING:
    case PADDING:
    {
      start = index;
      fieldSize = sizeOf((*iter).elementType_) * (*iter).length_;
      index += fieldSize;
      if(size < index)
      {
        throw SizeException("Insufficient data for adaptaion");
      }

      (*iter).field_.assign(&buffer[index], &buffer[index + fieldSize]);
      break;
    }
    case VARARRAY:
    case VARSTRING:
    {
      // Get the value of the length field
      NameTable::iterator length = nameTable_.find((*iter).varName_);
      if(length == nameTable_.end())
      {
        throw UndefinedFieldException("The variable length field identified by the specified name is undefined");
      }

      (*iter).length_ = getVarLength(length->second);

      start = index;
      fieldSize = sizeOf((*iter).elementType_) * (*iter).length_;
      index += fieldSize;
      if(size < index)
      {
        throw SizeException("Insufficient data for adaptaion");
      }

      (*iter).field_.assign(&buffer[index], &buffer[index + fieldSize]);
      break;
    }
    case TERMARRAY:
    case TERMSTRING:
    {
      (*iter).length_ = advanceToTerminator((*iter).elementType_, &buffer[index], size - index, (*iter).sentinal_);

      start = index;
      fieldSize = sizeOf((*iter).elementType_) * (*iter).length_;
      index += fieldSize;
      if(size < index)
      {
        throw SizeException("Insufficient data for adaptaion");
      }

      (*iter).field_.assign(&buffer[index], &buffer[index + fieldSize]);

      // Skip the terminator
      index += sizeOf((*iter).elementType_);
      break;
    }
    default:
      throw InvalidTypeException("Unrecognized field type");
    }
  }
}

// Fill buffer from format
void BinaryAdapterFormat::fill(char* buffer, size_t size) const
{
  size_t start;
  size_t index = 0;

  // Copy the filed contents to the buffer
  for(Fields::const_iterator iter = fields_.begin(); iter != fields_.end(); ++iter)
  {
    switch((*iter).fieldType_)
    {
    case ELEMENT:
    case VARARRAY:
    case VARSTRING:
    {
      // Append the contents of the field value to the buffer
      start = index;
      index += (*iter).field_.size();

      if(size < index)
      {
        throw SizeException("Insufficient buffer size");
      }

      std::copy((*iter).field_.begin(), (*iter).field_.end(), &buffer[start]);

      break;
    }
    case ARRAY:
    case STRING:
    case PADDING:
    {
      // Append the contents of the field value to the buffer
      // The full length specified for the field must be represented
      start = index;
      index += (*iter).length_ * sizeOf((*iter).elementType_);

      if(size < index)
      {
        throw SizeException("Insufficient buffer size");
      }

      std::copy((*iter).field_.begin(), (*iter).field_.end(), &buffer[start]);

      // Pad with zeroes if necessary
      size_t actual = start + (*iter).field_.size();
      std::fill(&buffer[actual], &buffer[index], 0);

      break;
    }
    case TERMARRAY:
    case TERMSTRING:
    {
      start = index;
      index += (*iter).field_.size() + sizeOf((*iter).elementType_);  // Length of data plus terminator

      if(size < index)
      {
        throw SizeException("Insufficient data for adaptaion");
      }

      std::copy((*iter).field_.begin(), (*iter).field_.end(), &buffer[start]);

      // Add the terminator
      start += (*iter).field_.size();
      appendTerminator((*iter).elementType_, &buffer[start], (*iter).sentinal_);

      break;
    }
    default:
      throw InvalidTypeException("Unrecognized field type");
    }
  }
}

// Clear all fields and adapted data
void BinaryAdapterFormat::clear()
{
  nameTable_.clear();
  fields_.clear();
}

size_t BinaryAdapterFormat::getDataSize() const
{
  size_t total = 0;
  for(Fields::const_iterator iter = fields_.begin(); iter != fields_.end(); ++iter)
  {
    size_t size = sizeOf((*iter).elementType_);
    total += (*iter).length_ * size;

    if((*iter).fieldType_ == TERMARRAY || (*iter).fieldType_ == TERMSTRING)
    {
      total += size;
    }
  }

  return total;
}

// Get the number of elements in the field with the specified name
size_t BinaryAdapterFormat::getFieldLength(const std::string& name) const
{
  NameTable::const_iterator iter = nameTable_.find(name);

  if(iter == nameTable_.end())
  {
    throw UndefinedFieldException("The field identified by the specified name is undefined");
  }

  return getFieldLength(iter->second);
}

// Get the number of bytes contained by the field with the specified name
size_t BinaryAdapterFormat::getFieldSize(const std::string& name) const
{
  NameTable::const_iterator iter = nameTable_.find(name);

  if(iter == nameTable_.end())
  {
    throw UndefinedFieldException("The field identified by the specified name is undefined");
  }

  return getFieldSize(iter->second);
}

// Get the field type for the field at the specified name
BinaryAdapterFormat::FieldType BinaryAdapterFormat::getFieldType(const std::string& name) const
{
  NameTable::const_iterator iter = nameTable_.find(name);

  if(iter == nameTable_.end())
  {
    throw UndefinedFieldException("The field identified by the specified name is undefined");
  }

  return getFieldType(iter->second);
}

// Get the element type for the field at the specified name
BinaryAdapterFormat::ElementType BinaryAdapterFormat::getElementType(const std::string& name) const
{
  NameTable::const_iterator iter = nameTable_.find(name);

  if(iter == nameTable_.end())
  {
    throw UndefinedFieldException("The field identified by the specified name is undefined");
  }

  return getElementType(iter->second);
}

// Get the element field value
template <typename T, BinaryAdapterFormat::ElementType E>
void BinaryAdapterFormat::getElementValue(size_t pos, T& value) const
{
  if(fields_[pos].fieldType_ != ELEMENT)
  {
    throw InvalidTypeException("Attempting to extract a non-element field with an element type");
  }

  if(fields_[pos].elementType_ != E)
  {
    throw InvalidTypeException("Value type for extraction does not match element type for field");
  }

  const char* start = &(fields_[pos].field_[0]);
  std::copy(start, start + sizeof(T), reinterpret_cast<char*>(&value));

  if(order_ != SYSTEMBYTEORDER)
  {
    swap_bytes(value);
  }
}

// Get the array field value
template <typename T, BinaryAdapterFormat::ElementType E>
void BinaryAdapterFormat::getArrayValue(size_t pos, std::vector<T>& value) const
{
  if(fields_[pos].fieldType_ != ARRAY && fields_[pos].fieldType_ != VARARRAY && fields_[pos].fieldType_ != TERMARRAY && fields_[pos].fieldType_ != PADDING)
  {
    throw InvalidTypeException("Attempting to extract an element field with a non-element type");
  }

  if(fields_[pos].elementType_ != E)
  {
    throw InvalidTypeException("Value type for extraction does not match element type for field");
  }

  const T* start = reinterpret_cast<const T*>(&(fields_[pos].field_[0]));
  value.assign(&start[0], &start[fields_[pos].length_]);

  if(order_ != SYSTEMBYTEORDER)
  {
    swap_bytes_array(&value[0], value.size());
  }
}

// Get the field value from field name
template <typename T>
void BinaryAdapterFormat::getNamedValue(const std::string& name, T& value) const
{
  NameTable::const_iterator iter = nameTable_.find(name);

  if(iter == nameTable_.end())
  {
    throw UndefinedFieldException("The field identified by the specified name is undefined");
  }

  getFieldValue(iter->second, value);
}

// Get the string field value
void BinaryAdapterFormat::getFieldValue(size_t pos, std::string& value) const
{
  if(fields_[pos].fieldType_ != STRING && fields_[pos].fieldType_ != VARSTRING && fields_[pos].fieldType_ != TERMSTRING)
  {
    throw InvalidTypeException("Attempting to extract a string field with a non-string type");
  }

  const char* start = &(fields_[pos].field_[0]);
  value.assign(&start[0], &start[fields_[pos].length_]);
}

// Set the element field value
template <typename T, BinaryAdapterFormat::ElementType E>
void BinaryAdapterFormat::setElementValue(size_t pos, T value)
{
  if(fields_[pos].fieldType_ != ELEMENT)
  {
    throw InvalidTypeException("Attempting to assign a non-element field with an element type");
  }

  if(fields_[pos].elementType_ != E)
  {
    throw InvalidTypeException("Value type for assignment does not match element type for field");
  }

  if(order_ != SYSTEMBYTEORDER)
  {
    swap_bytes(value);
  }

  const char* start = reinterpret_cast<const char*>(&value);
  fields_[pos].field_.assign(start, start + sizeof(T));
}

// Set the array field value
template <typename T, BinaryAdapterFormat::ElementType E>
void BinaryAdapterFormat::setArrayValue(size_t pos, const std::vector<T>& value)
{
  if(fields_[pos].fieldType_ != ARRAY && fields_[pos].fieldType_ != VARARRAY && fields_[pos].fieldType_ != TERMARRAY && fields_[pos].fieldType_ != PADDING)
  {
    throw InvalidTypeException("Attempting to assign a non-array field with an array type");
  }

  if(fields_[pos].elementType_ != E)
  {
    throw InvalidTypeException("Value type for assignment does not match element type for field");
  }

  // Validate array lengths; treat padding as a fixed length array
  if((fields_[pos].fieldType_ == ARRAY || fields_[pos].fieldType_ == PADDING) && (fields_[pos].length_ < value.size()))
  {
    throw SizeException("Value exceeds specified array field length");
  }
  else
  {
    // Set the length for the variable types
    fields_[pos].length_ = value.size();

    // Set the length field value
    if(fields_[pos].fieldType_ == VARARRAY)
    {
      NameTable::iterator length = nameTable_.find(fields_[pos].varName_);
      if(length == nameTable_.end())
      {
        throw UndefinedFieldException("The variable length field identified by the specified name is undefined");
      }

      setVarLength(length->second, value.size());
    }
  }

  const char* start = reinterpret_cast<const char*>(&value[0]);
  fields_[pos].field_.assign(start, start + (sizeof(T) * value.size()));

  if(order_ != SYSTEMBYTEORDER)
  {
    T* swap = reinterpret_cast<T*>(&(fields_[pos].field_[0]));
    swap_bytes_array(&swap[0], value.size());
  }
}

// Set the field value from field name
template <typename T>
void BinaryAdapterFormat::setNamedValue(const std::string& name, T value)
{
  NameTable::const_iterator iter = nameTable_.find(name);

  if(iter == nameTable_.end())
  {
    throw UndefinedFieldException("The field identified by the specified name is undefined");
  }

  setFieldValue(iter->second, value);
}

// Set the string field value
void BinaryAdapterFormat::setFieldValue(size_t pos, const std::string& value)
{
  if(fields_[pos].fieldType_ != STRING && fields_[pos].fieldType_ != VARSTRING && fields_[pos].fieldType_ != TERMSTRING)
  {
    throw InvalidTypeException("Attempting to assign a non-string field with a string type");
  }

  // Validate string length
  if((fields_[pos].fieldType_ == STRING) && (fields_[pos].length_ < value.size()))
  {
    throw SizeException("Value exceeds specified string field length");
  }
  else
  {
    // Set the length for the variable types
    fields_[pos].length_ = value.size();

    // Set the length field value
    if(fields_[pos].fieldType_ == VARSTRING)
    {
      NameTable::iterator length = nameTable_.find(fields_[pos].varName_);
      if(length == nameTable_.end())
      {
        throw UndefinedFieldException("The variable length field identified by the specified name is undefined");
      }

      setVarLength(length->second, value.size());
    }
  }

  fields_[pos].field_.assign(value.begin(), value.end());
}

void BinaryAdapterFormat::setFieldValue(const std::string& name, const std::string& value)
{
  NameTable::const_iterator iter = nameTable_.find(name);

  if(iter == nameTable_.end())
  {
    throw UndefinedFieldException("The field identified by the specified name is undefined");
  }

  setFieldValue(iter->second, value);
}
