#ifndef BINARYADAPTERFORMAT_H
#define BINARYADAPTERFORMAT_H

#include <inttypes.h>
#include <string>
#include <vector>
#include <map>

#include "byteorder.h"
#include "Exceptions.h"
#include "PoolAllocator.h"
#include "Format.h"

namespace phylum {

// A collection of memory adjacent field
// A field may consist of one or more basic data elements
// Each element has a type, size, and value
class BinaryAdapterFormat : public Format
{
public:
  enum ElementType
  {
    INT8,
    UINT8,
    INT16,
    UINT16,
    INT32,
    UINT32,
    INT64,
    UINT64,
    FLOAT32,
    FLOAT64
  };

  enum FieldType
  {
    ELEMENT,
    ARRAY,        // Fixed length array, length is pre-defined
    VARARRAY,     // Variable length array, length specified by a separate field
    TERMARRAY,    // Variable length array, terminated with a specific value
    STRING,       // Fixed length string, with pre-defined length
    VARSTRING,    // Variable length string, with length specified by a separate field
    TERMSTRING,   // Variable length string, with terminator
    PADDING       // Add a padding field
  };

protected:
  class Field
  {
  public:
    typedef std::vector<char> Value;

    FieldType fieldType_;

    ElementType elementType_;   // Indicates type and size

    size_t length_;             // Length of an array field

    std::string varName_;       // Name of element specifying the vararray length

    double sentinal_;           // Terminator value for termarray

    Value field_;               // Start of field within the buffer

  public:
    Field(FieldType fieldType, ElementType elementType) : fieldType_(fieldType), elementType_(elementType), length_(1) { }

    Field(FieldType fieldType, ElementType elementType, size_t arrayLength) : fieldType_(fieldType), elementType_(elementType), length_(arrayLength) { }

    Field(FieldType fieldType, ElementType elementType, const std::string& varName) : fieldType_(fieldType), elementType_(elementType), length_(0), varName_(varName) { }

    Field(FieldType fieldType, ElementType elementType, double sentinal) : fieldType_(fieldType), elementType_(elementType), length_(0), sentinal_(sentinal) { }

    Field(const Field& copy) : fieldType_(copy.fieldType_), elementType_(copy.elementType_), length_(copy.length_), varName_(copy.varName_), sentinal_(copy.sentinal_), field_(copy.field_.begin(), copy.field_.end()) { }
  };

  typedef std::vector<Field> Fields;

  typedef std::map<std::string, size_t> NameTable;

protected:
  ByteOrder order_;           // Big or little endian byte order

  bool networkOrder_;         // Is network byte order?

  Fields fields_;             // Fields of format to adapt

  NameTable nameTable_;       // Name to field index lookup

protected:
  size_t sizeOf(ElementType type) const;

  // For getVarArrayLength
  template <typename T>
  size_t getLength(size_t pos) const
  {
    T value;
    getFieldValue(pos, value);
    return static_cast<size_t>(value);
  }

  size_t getVarLength(size_t pos) const;

  void setVarLength(size_t pos, size_t length);

  // For advanceToTerminator, counts number of items in array not including the sentinal value
  template <typename T>
  size_t advanceTo(const char* buffer, size_t length, double sentinal) const
  {
    const T* data = reinterpret_cast<const T*>(buffer);
    size_t actual = length / sizeof(T);

    for(size_t i = 0; i < actual; i++)
    {
      if(data[i] == static_cast<T>(sentinal))
      {
        return (i * sizeof(T));
      }
    }

    throw UnterminatedFieldException("End of data buffer was encountered before the specified terminator was found");
  }

  size_t advanceToTerminator(ElementType type, const char* buffer, size_t len, double sentinal) const;

  // For appendTerminator
  template <typename T>
  void appendTo(char* buffer, double sentinal) const
  {
    T value = static_cast<T>(sentinal);
    T* array = reinterpret_cast<T*>(buffer);
    *array = value;
  }

  void appendTerminator(ElementType type, char* buffer, double sentinal) const;

  template <typename T, ElementType E>
  void getElementValue(size_t pos, T& value) const;

  template <typename T, ElementType E>
  void getArrayValue(size_t pos, std::vector<T>& value) const;

  // Get the field value from field name
  template <typename T>
  void getNamedValue(const std::string& name, T& value) const;

  template <typename T, ElementType E>
  void setElementValue(size_t pos, T value);

  template <typename T, ElementType E>
  void setArrayValue(size_t pos, const std::vector<T>& value);

  // Set the field value from field name
  template <typename T>
  void setNamedValue(const std::string& name, T value);

public:
  BinaryAdapterFormat(unsigned int type, ByteOrder order, Creator* creator) : Format(type, creator), order_(order), networkOrder_(order == ORDER_BIG) { }

  BinaryAdapterFormat(unsigned int type, bool isNetworkByteOrder, Creator* creator) : Format(type, creator), networkOrder_(isNetworkByteOrder) { order_ = isNetworkByteOrder ? ORDER_BIG : ORDER_LITTLE; }

  BinaryAdapterFormat(const BinaryAdapterFormat& copy, Creator* creator) : Format(copy, creator), order_(copy.order_), networkOrder_(copy.networkOrder_), fields_(copy.fields_.begin(), copy.fields_.end()), nameTable_(copy.nameTable_.begin(), copy.nameTable_.end()) { }

  BinaryAdapterFormat* clone(Creator* creator) const { return new BinaryAdapterFormat(*this, creator); }

  POOL_DECLARE(BinaryAdapterFormat)

  // Add a single element
  void addField(const std::string& name, ElementType type);

  // Add an array of elements
  void addArrayField(const std::string& name, ElementType type, size_t length);

  // Add a variable length array of elements
  void addVarArrayField(const std::string& name, ElementType type, const std::string& elementName);

  // Add a terminated variable length array of elements
  void addTermArrayField(const std::string& name, ElementType type, double terminator);

  // Add a fixed length string
  void addStringField(const std::string& name, size_t length);

  // Add a variable length string
  void addVarStringField(const std::string& name, const std::string& elementName);

  // Add a terminated variable length string
  void addTermStringField(const std::string& name, double terminator);

  // Add a padding field of one or more bytes
  void addPaddingField(size_t size);

  // Add a named padding field; experience shows that some times values are hidden in padding
  void addPaddingField(const std::string& name, size_t size);

  // Adapt buffer to format
  void adapt(const char* buffer, size_t size);

  // Fill buffer from format
  void fill(char* buffer, size_t size) const;

  // Clear all fields and adapted data
  void clear();

  // Get the number of bytes contained in the adapter
  size_t getDataSize() const;

  // Get the number of fields in the format
  size_t getNumFields() const { return fields_.size(); }

  // Get the number of elements in the field at the specified position
  size_t getFieldLength(size_t pos) const { return fields_[pos].length_; }

  // Get the number of elements in the field with the specified name
  size_t getFieldLength(const std::string& name) const;

  // Get the number of bytes contained by the field at the specified position
  size_t getFieldSize(size_t pos) const { return fields_[pos].length_ * sizeOf(fields_[pos].elementType_); }

  // Get the number of bytes contained by the field with the specified name
  size_t getFieldSize(const std::string& name) const;

  // Get the field type for the field at the specified position
  FieldType getFieldType(size_t pos) const { return fields_[pos].fieldType_; }

  // Get the field type for the field at the specified name
  FieldType getFieldType(const std::string& name) const;

  // Get the element type for the field at the specified position
  ElementType getElementType(size_t pos) const { return fields_[pos].elementType_; }

  // Get the element type for the field at the specified name
  ElementType getElementType(const std::string& name) const;

  // Get the int8 value
  void getFieldValue(size_t pos, int8_t& value) const { getElementValue<int8_t, INT8>(pos, value); }

  // Get the int8 value
  void getFieldValue(const std::string& name, int8_t& value) const { getNamedValue(name, value); }

  // Get the uint8 value
  void getFieldValue(size_t pos, uint8_t& value) const { getElementValue<uint8_t, UINT8>(pos, value); }

  // Get the uint8 value
  void getFieldValue(const std::string& name, uint8_t& value) const { getNamedValue(name, value); }

  // Get the int16 value
  void getFieldValue(size_t pos, int16_t& value) const { getElementValue<int16_t, INT16>(pos, value); }

  // Get the int16 value
  void getFieldValue(const std::string& name, int16_t& value) const { getNamedValue(name, value); }

  // Get the uint16 value
  void getFieldValue(size_t pos, uint16_t& value) const { getElementValue<uint16_t, UINT16>(pos, value); }

  // Get the uint16 value
  void getFieldValue(const std::string& name, uint16_t& value) const { getNamedValue(name, value); }

  // Get the int32 value
  void getFieldValue(size_t pos, int32_t& value) const { getElementValue<int32_t, INT32>(pos, value); }

  // Get the int32 value
  void getFieldValue(const std::string& name, int32_t& value) const { getNamedValue(name, value); }

  // Get the uint32 value
  void getFieldValue(size_t pos, uint32_t& value) const { getElementValue<uint32_t, UINT32>(pos, value); }

  // Get the uint32 value
  void getFieldValue(const std::string& name, uint32_t& value) const { getNamedValue(name, value); }

  // Get the int64 value
  void getFieldValue(size_t pos, int64_t& value) const { getElementValue<int64_t, INT64>(pos, value); }

  // Get the int64 value
  void getFieldValue(const std::string& name, int64_t& value) const { getNamedValue(name, value); }

  // Get the uint64 value
  void getFieldValue(size_t pos, uint64_t& value) const { getElementValue<uint64_t, UINT64>(pos, value); }

  // Get the uint64 value
  void getFieldValue(const std::string& name, uint64_t& value) const { getNamedValue(name, value); }

  // Get the float value
  void getFieldValue(size_t pos, float& value) const { getElementValue<float, FLOAT32>(pos, value); }

  // Get the float value
  void getFieldValue(const std::string& name, float& value) const { getNamedValue(name, value); }

  // Get the double value
  void getFieldValue(size_t pos, double& value) const { getElementValue<double, FLOAT64>(pos, value); }

  // Get the double value
  void getFieldValue(const std::string& name, double& value) const { getNamedValue(name, value); }

  // Get the int8 array field value
  void getFieldValue(size_t pos, std::vector<int8_t>& value) const { getArrayValue<int8_t, INT8>(pos, value); }

  // Get the int8 array field value
  void getFieldValue(const std::string& name, std::vector<int8_t>& value) const { getNamedValue(name, value); }

  // Get the uint8 array field value
  void getFieldValue(size_t pos, std::vector<uint8_t>& value) const { getArrayValue<uint8_t, UINT8>(pos, value); }

  // Get the uint8 array field value
  void getFieldValue(const std::string& name, std::vector<uint8_t>& value) const { getNamedValue(name, value); }

  // Get the int16 array field value
  void getFieldValue(size_t pos, std::vector<int16_t>& value) const { getArrayValue<int16_t, INT16>(pos, value); }

  // Get the int16 array field value
  void getFieldValue(const std::string& name, std::vector<int16_t>& value) const { getNamedValue(name, value); }

  // Get the uint16 array field value
  void getFieldValue(size_t pos, std::vector<uint16_t>& value) const { getArrayValue<uint16_t, UINT16>(pos, value); }

  // Get the uint16 array field value
  void getFieldValue(const std::string& name, std::vector<uint16_t>& value) const { getNamedValue(name, value); }

  // Get the int32 array field value
  void getFieldValue(size_t pos, std::vector<int32_t>& value) const { getArrayValue<int32_t, INT32>(pos, value); }

  // Get the int32 array field value
  void getFieldValue(const std::string& name, std::vector<int32_t>& value) const { getNamedValue(name, value); }

  // Get the uint32 array field value
  void getFieldValue(size_t pos, std::vector<uint32_t>& value) const { getArrayValue<uint32_t, UINT32>(pos, value); }

  // Get the uint32 array field value
  void getFieldValue(const std::string& name, std::vector<uint32_t>& value) const { getNamedValue(name, value); }

  // Get the int64 array field value
  void getFieldValue(size_t pos, std::vector<int64_t>& value) const { getArrayValue<int64_t, INT64>(pos, value); }

  // Get the int64 array field value
  void getFieldValue(const std::string& name, std::vector<int64_t>& value) const { getNamedValue(name, value); }

  // Get the uint64 array field value
  void getFieldValue(size_t pos, std::vector<uint64_t>& value) const { getArrayValue<uint64_t, UINT64>(pos, value); }

  // Get the uint64 array field value
  void getFieldValue(const std::string& name, std::vector<uint64_t>& value) const { getNamedValue(name, value); }

  // Get the float array field value
  void getFieldValue(size_t pos, std::vector<float>& value) const { getArrayValue<float, FLOAT32>(pos, value); }

  // Get the float array field value
  void getFieldValue(const std::string& name, std::vector<float>& value) const { getNamedValue(name, value); }

  // Get the double array field value
  void getFieldValue(size_t pos, std::vector<double>& value) const { getArrayValue<double, FLOAT64>(pos, value); }

  // Get the double array field value
  void getFieldValue(const std::string& name, std::vector<double>& value) const { getNamedValue(name, value); }

  // Get the string field value
  void getFieldValue(size_t pos, std::string& value) const;

  // Get the string field value
  void getFieldValue(const std::string& name, std::string& value) const { getNamedValue(name, value); }

  // Set the int8 value
  void setFieldValue(size_t pos, int8_t value) { setElementValue<int8_t, INT8>(pos, value); }

  // Set the int8 value
  void setFieldValue(const std::string& name, int8_t value) { setNamedValue(name, value); }

  // Set the uint8 value
  void setFieldValue(size_t pos, uint8_t value) { setElementValue<uint8_t, UINT8>(pos, value); }

  // Set the uint8 value
  void setFieldValue(const std::string& name, uint8_t value) { setNamedValue(name, value); }

  // Set the int16 value
  void setFieldValue(size_t pos, int16_t value) { setElementValue<int16_t, INT16>(pos, value); }

  // Set the int16 value
  void setFieldValue(const std::string& name, int16_t value) { setNamedValue(name, value); }

  // Set the uint16 value
  void setFieldValue(size_t pos, uint16_t value) { setElementValue<uint16_t, UINT16>(pos, value); }

  // Set the uint16 value
  void setFieldValue(const std::string& name, uint16_t value) { setNamedValue(name, value); }

  // Set the int32 value
  void setFieldValue(size_t pos, int32_t value) { setElementValue<int32_t, INT32>(pos, value); }

  // Set the int32 value
  void setFieldValue(const std::string& name, int32_t value) { setNamedValue(name, value); }

  // Set the uint32 value
  void setFieldValue(size_t pos, uint32_t value) { setElementValue<uint32_t, UINT32>(pos, value); }

  // Set the uint32 value
  void setFieldValue(const std::string& name, uint32_t value) { setNamedValue(name, value); }

  // Set the int64 value
  void setFieldValue(size_t pos, int64_t value) { setElementValue<int64_t, INT64>(pos, value); }

  // Set the int64 value
  void setFieldValue(const std::string& name, int64_t value) { setNamedValue(name, value); }

  // Set the uint64 value
  void setFieldValue(size_t pos, uint64_t value) { setElementValue<uint64_t, UINT64>(pos, value); }

  // Set the uint64 value
  void setFieldValue(const std::string& name, uint64_t value) { setNamedValue(name, value); }

  // Set the float value
  void setFieldValue(size_t pos, float value) { setElementValue<float, FLOAT32>(pos, value); }

  // Set the float value
  void setFieldValue(const std::string& name, float value) { setNamedValue(name, value); }

  // Set the double value
  void setFieldValue(size_t pos, double value) { setElementValue<double, FLOAT64>(pos, value); }

  // Set the double value
  void setFieldValue(const std::string& name, double value) { setNamedValue(name, value); }

  // Set the int8 array field value
  void setFieldValue(size_t pos, const std::vector<int8_t>& value) { setArrayValue<int8_t, INT8>(pos, value); }

  // Set the int8 array field value
  void setFieldValue(const std::string& name, const std::vector<int8_t>& value) { setNamedValue(name, value); }

  // Set the uint8 array field value
  void setFieldValue(size_t pos, const std::vector<uint8_t>& value) { setArrayValue<uint8_t, UINT8>(pos, value); }

  // Set the uint8 array field value
  void setFieldValue(const std::string& name, const std::vector<uint8_t>& value) { setNamedValue(name, value); }

  // Set the int16 array field value
  void setFieldValue(size_t pos, const std::vector<int16_t>& value) { setArrayValue<int16_t, INT16>(pos, value); }

  // Set the int16 array field value
  void setFieldValue(const std::string& name, const std::vector<int16_t>& value) { setNamedValue(name, value); }

  // Set the uint16 array field value
  void setFieldValue(size_t pos, const std::vector<uint16_t>& value) { setArrayValue<uint16_t, UINT16>(pos, value); }

  // Set the uint16 array field value
  void setFieldValue(const std::string& name, const std::vector<uint16_t>& value) { setNamedValue(name, value); }

  // Set the int32 array field value
  void setFieldValue(size_t pos, const std::vector<int32_t>& value) { setArrayValue<int32_t, INT32>(pos, value); }

  // Set the int32 array field value
  void setFieldValue(const std::string& name, const std::vector<int32_t>& value) { setNamedValue(name, value); }

  // Set the uint32 array field value
  void setFieldValue(size_t pos, const std::vector<uint32_t>& value) { setArrayValue<uint32_t, UINT32>(pos, value); }

  // Set the uint32 array field value
  void setFieldValue(const std::string& name, const std::vector<uint32_t>& value) { setNamedValue(name, value); }

  // Set the int64 array field value
  void setFieldValue(size_t pos, const std::vector<int64_t>& value) { setArrayValue<int64_t, INT64>(pos, value); }

  // Set the int64 array field value
  void setFieldValue(const std::string& name, const std::vector<int64_t>& value) { setNamedValue(name, value); }

  // Set the uint64 array field value
  void setFieldValue(size_t pos, const std::vector<uint64_t>& value) { setArrayValue<uint64_t, UINT64>(pos, value); }

  // Set the uint64 array field value
  void setFieldValue(const std::string& name, const std::vector<uint64_t>& value) { setNamedValue(name, value); }

  // Set the float array field value
  void setFieldValue(size_t pos, const std::vector<float>& value) { setArrayValue<float, FLOAT32>(pos, value); }

  // Set the float array field value
  void setFieldValue(const std::string& name, const std::vector<float>& value) { setNamedValue(name, value); }

  // Set the double array field value
  void setFieldValue(size_t pos, const std::vector<double>& value) { setArrayValue<double, FLOAT64>(pos, value); }

  // Set the double array field value
  void setFieldValue(const std::string& name, const std::vector<double>& value) { setNamedValue(name, value); }

  // Set the string field value
  void setFieldValue(size_t pos, const std::string& value);

  // Set the string field value
  void setFieldValue(const std::string& name, const std::string& value);
};

} // namespace phylum

#endif // BINARYADAPTERFORMAT_H
