#ifndef CACHE_H
#define CACHE_H

#include <tbb/spin_rw_mutex.h>
#include "Creator.h"
#include "Exceptions.h"
#include "FormatVisitor.h"

class Format;

class Cache : public Creator
{
protected:
  /// Get an item from the cache
  virtual Format* get(size_t pos) = 0;

  /// Resize the cache if it exceeds the max size
  virtual void resize(size_t maxsize) = 0;

protected:
  typedef tbb::spin_rw_mutex CacheMutex;

  mutable CacheMutex mutex_;

  size_t maxsize_;

public:
  class CacheAccessor;

  class ConstCacheAccessor;

  Cache(size_t maxsize) : maxsize_(maxsize) { }

  /// Destructor
  virtual ~Cache() { }

  /// Get maximum size of the cache
  size_t getMaxSize() const
  {
    size_t maxsize;

    { // Lock
    CacheMutex::scoped_lock lock(mutex_, /*is_writer=*/false);
    maxsize = maxsize_;
    } // Unlock

    return maxsize;
  }

  /// Set the maximum size of the cache
  void setMaxSize(size_t maxsize)
  {
    CacheMutex::scoped_lock lock(mutex_, /*is_writer=*/true);

    if(maxsize_ > maxsize)
    {
      resize(maxsize);
    }

    maxsize_ = maxsize;
  }

  /// Get the current size of the cache
  virtual size_t getSize() const = 0;

  /// Get an item from the cache with a write lock
  void getData(CacheAccessor& accessor, size_t pos)
  {
    accessor.acquire(mutex_);
    accessor.setFormat(get(pos));
  }

  /// Get an item from the cache with a read lock
  void getData(ConstCacheAccessor& accessor, size_t pos)
  {
    accessor.acquire(mutex_);
    accessor.setFormat(get(pos));
  }

  /// Add a message to the cache
  virtual void addData(const Format* fmt) = 0;

  /// Visit items in the cache
  virtual void visit(FormatVisitor& visitor) { throw NotImplementedException("The visit functionality must be implemented in a subclass"); }

  /// Visit items in the cache
  virtual void visit(ConstFormatVisitor& visitor) const { throw NotImplementedException("The visit functionality must be implemented in a subclass"); }

  // Concurrent cache accessor; uses a write lock
  class CacheAccessor
  {
  protected:
    Format* format_;

    bool locked_;

  public:
    CacheAccessor() : format_(NULL), locked_(false) { }

    ~CacheAccessor() { release(); }

    void release()
    {
      if(locked_)
      {
        format_ = NULL;
        lock_.release();
        locked_ = false;
      }
    }

    Format* getFormat() { return format_; }

  private:
    friend class Cache;

    CacheMutex::scoped_lock lock_;

    void acquire(CacheMutex& mutex)
    {
      lock_.acquire(mutex, /*is_writer=*/true);
      locked_ = true;
    }

    void setFormat(Format* fmt) { format_ = fmt; }
  };

  // Constant Concurrent cache accessor; uses a read lock
  class ConstCacheAccessor
  {
  protected:
    Format* format_;

    bool locked_;

  public:
    ConstCacheAccessor() : format_(NULL), locked_(false) { }

    ~ConstCacheAccessor() { release(); }

    void release()
    {
      if(locked_)
      {
        format_ = NULL;
        lock_.release();
        locked_ = false;
      }
    }

    const Format* getFormat() const { return format_; }

  private:
    friend class Cache;

    CacheMutex::scoped_lock lock_;

    void acquire(CacheMutex& mutex)
    {
      lock_.acquire(mutex, /*is_writer=*/true);
      locked_ = true;
    }

    void setFormat(Format* fmt) { format_ = fmt; }
  };
};

#endif // CACHE_H
