#ifdef WIN32
#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif
#include <io.h>
#include <windows.h>
#else
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif

#include "RawFormat.h"
#include "Exceptions.h"
#include "FileSource.h"

// IN, OUT, and INOUT constructor
FileSource::FileSource(Mode mode, const std::string& filename)
: Source(mode),
  filename_(filename),
  blockSize_(RawFormat::FILE_BUFFER)
{
#ifdef WIN32
  win32event_ = INVALID_HANDLE;
#endif
  openFile(mode, filename, NONE);
}

// OUT and INOUT constructor with create or append option
FileSource::FileSource(Mode mode, const std::string& filename, unsigned int filemode)
: Source(mode),
  filename_(filename),
  blockSize_(RawFormat::FILE_BUFFER)
{
#ifdef WIN32
  win32event_ = INVALID_HANDLE;
#endif
  openFile(mode, filename, filemode);
}

// IN, OUT, and INOUT constructor
FileSource::FileSource(Mode mode, unsigned int blockSize, const std::string& filename)
: Source(mode),
  filename_(filename),
  blockSize_(blockSize)
{
#ifdef WIN32
  win32event_ = INVALID_HANDLE;
#endif
  openFile(mode, filename, NONE);
}

// OUT and INOUT constructor with create or append option
FileSource::FileSource(Mode mode, unsigned int blockSize, const std::string& filename, unsigned int filemode)
: Source(mode),
  filename_(filename),
  blockSize_(blockSize)
{
#ifdef WIN32
  win32event_ = INVALID_HANDLE;
#endif
  openFile(mode, filename, filemode);
}

FileSource::~FileSource()
{
  if(fd_ != -1)
  {
    close();
  }
}

void FileSource::openFile(Mode mode, const std::string& filename, unsigned int filemode)
{
  int flags = 0;
  int perm = S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH;

  switch(mode)
  {
    case IN:
      flags = O_RDONLY;
      break;
    case OUT:
      flags = O_WRONLY;
      break;
    case INOUT:
      flags = O_RDWR;
      break;
    default:
      state_ = ERROR;
      throw UnsupportedModeException("The specified file i/o mode is not supported");
      break;
  }

  if(filemode & CREAT)
  {
    flags |= O_CREAT;
  }

  if(filemode & EXCL)
  {
    flags |= O_EXCL;
  }

  if(filemode & TRUNC)
  {
    flags |= O_TRUNC;
  }

#ifdef WIN32
  if(filemode & BINARY)
  {
    flags |= O_BINARY;
  }
#endif

  fd_ = ::open(filename.c_str(), flags, perm);

  if(fd_ == -1)
  {
    state_ = ERROR;
    throw OpenFailedException("Failed to open file source");
  }

  // Set handles
#ifdef WIN32
  // Create an event that is always signaled to simulate
  // file always ready to read and write
  win32event_ = CreateEvent(NULL, TRUE, TRUE, NULL);
  Handle h = win32event_;
#else
  // Use file handle for UNIX
  Handle h = fd_;
#endif

  if(mode == IN || mode == INOUT)
  {
    readHandle_ = h;
  }

  if(mode == OUT || mode == INOUT)
  {
    writeHandle_ = h;
  }

  state_ = ACTIVE;
}

void FileSource::close()
{
  if(fd_ == -1)
  {
    throw NotOpenException("Attempt to close a file source which is not open");
  }

  ::close(fd_);
  fd_ = -1;
  readHandle_ = INVALID_HANDLE;
  writeHandle_ = INVALID_HANDLE;

#ifdef WIN32
  CloseHandle(win32event_);
  win32event_ = INVALID_HANDLE;
#endif

  state_ = INACTIVE;
}

Format* FileSource::readRaw()
{
  RawFormat* fmt = new RawFormat(blockSize_, this);

  int total = ::read(fd_, fmt->getBuffer(), fmt->getBufferSize());

  if(total <= 0)
  {
    delete fmt;

    if(total == 0)
    {
      state_ = DONE;
      throw EOFException("The end of the file has been reached");
    }
    else
    {
      state_ = ERROR;
      throw ReadFailedException("Error reading from file source");
    }
  }

  fmt->setDataSize(total);

  return fmt;
}

void FileSource::writeRaw(const Format* fmt)
{
  const RawFormat* rawfmt = dynamic_cast< const RawFormat* >(fmt);
  if(rawfmt != NULL)
  {
    if(rawfmt->getDataSize() > 0)
    {
      int total = ::write(fd_, rawfmt->getBuffer(), rawfmt->getDataSize());

      if(total < 0)
      {
        state_ = ERROR;
        throw WriteFailedException("Error writing to file source");
      }
    }
  }
  else
  {
    state_ = ERROR;
    throw UnsupportedFormatException("File source could not process the specified format type for writing");
  }
}
