#ifndef FORWARDTERMINATOR_H
#define FORWARDTERMINATOR_H

#include <algorithm>
#include <vector>
#include "PipelineFunctions.h"
#include "Pipeline.h"
#include "Source.h"

// Pipeline terminator responsible for forwarding data to an output
// source or a pipeline.  Allows pipelines to be chained together.
// A general way to forward pipeline data directly to an output source
// or through a pipeline preceeding an output source.
class ForwardTerminator : public Terminator
{
protected:
  typedef std::vector< Source* > SourceList;

  typedef std::vector< Pipeline* > PipelineList;

  SourceList sources_;

  PipelineList pipelines_;

public:
  // Constructor
  ForwardTerminator(bool is_serial) : Terminator(is_serial) { }

  size_t getNumSources() const { return sources_.size(); }

  size_t getNumPipelines() const { return pipelines_.size(); }

  void addSource(Source* source) { sources_.push_back(source); }

  void addPipeline(Pipeline* pipeline) { pipelines_.push_back(pipeline); }

  void removeSource(Source* source)
  {
    SourceList::iterator iter = std::find(sources_.begin(), sources_.end(), source);

    if(iter != sources_.end())
    {
      sources_.erase(iter);
    }
  }

  void removePipeline(Pipeline* pipeline)
  {
    PipelineList::iterator iter = std::find(pipelines_.begin(), pipelines_.end(), pipeline);

    if(iter != pipelines_.end())
    {
      pipelines_.erase(iter);
    }
  }

  Source* getSource(size_t pos) const { return sources_[pos]; }

  Pipeline* getPipeline(size_t pos) const { return pipelines_[pos]; }

  void clear() { sources_.clear(); pipelines_.clear(); }

  virtual bool operator()(Format* format)
  {
    for(SourceList::iterator iter = sources_.begin(); iter != sources_.end(); ++iter)
    {
      (*iter)->enqueueForWrite(format->clone(this));
    }

    for(PipelineList::iterator iter = pipelines_.begin(); iter != pipelines_.end(); ++iter)
    {
      (*iter)->addData(format->clone(this));
    }

    format->getCreator()->deallocate(format);

    return true;
  }
};

#endif // FORWARDTERMINATOR_H

