#include <algorithm>
#include <functional>
#include "Format.h"
#include "LRUCache.h"

/// Get an item from the cache
Format* LRUCache::get(size_t pos)
{
  if(pos < 0 || pos > cache_.size())
  {
    throw OutOfBoundsException("The specified index is out of bounds");
  }

  return cache_[pos];
}

/// Resize the cache if it exceeds the max size
void LRUCache::resize(size_t maxsize)
{
  if(maxsize > cache_.size())
  {
    cache_.resize(maxsize);
  }
}

/// Get the size of the cache
size_t LRUCache::getSize() const
{
  size_t size;

  { // Lock
  CacheMutex::scoped_lock lock(mutex_, /*is_writer=*/false);
  size = cache_.size();
  } // Unlock

  return size;
}

/// Add a message to the cache
void LRUCache::addData(const Format* fmt)
{
  CacheMutex::scoped_lock lock(mutex_, /*is_writer=*/true);

  while(cache_.size() >= maxsize_)
  {
    Format* front = cache_.front();
    cache_.pop_front();
    front->getCreator()->deallocate(front);
  }

  Format* copy = fmt->clone(this);
  cache_.push_back(copy);
}

/// Visit items in the cache
void LRUCache::visit(FormatVisitor& visitor)
{
  CacheMutex::scoped_lock lock(mutex_, /*is_writer=*/true);
  for(FormatCache::iterator iter = cache_.begin(); iter != cache_.end(); ++iter)
  {
    (*iter)->accept(visitor);
  }
}

/// Visit items in the cache
void LRUCache::visit(ConstFormatVisitor& visitor) const
{
  CacheMutex::scoped_lock lock(mutex_, /*is_writer=*/false);
  for(FormatCache::const_iterator iter = cache_.begin(); iter != cache_.end(); ++iter)
  {
    (*iter)->accept(visitor);
  }
}
