#include <functional>
#include "Format.h"
#include "LRUMultiCache.h"

/// Get an item from the cache associated with the specified ID
template <typename ID, typename GetID>
Format* LRUMultiCache<ID, GetID>::get(size_t pos)
{
  typename MultiCache::iterator iter = cache_.find(queryCache_);

  if(iter == cache_.end())
  {
    throw NoEntryExistsException("A cache entry for the selected query cache ID does not exist");
  }

  if(pos < 0 || pos > iter->second.size())
  {
    throw OutOfBoundsException("The specified index is out of bounds");
  }

  return (*iter)[pos];
}

/// Resize the cache if it exceeds the max size
template <typename ID, typename GetID>
void LRUMultiCache<ID, GetID>::resize(size_t maxsize)
{
//  std::for_each(cache_.begin(), cache_.end(), std::compose1(std::bind1st(std::mem_fun(&FormatCache::resize), size), select2nd<MultiCache::value_type>()));
  for(typename MultiCache::iterator iter = cache_.begin(); iter != cache_.end(); ++iter)
  {
    if(iter->second.size() > maxsize)
    {
      iter->second.resize();
    }
  }
}

/// Set the id of the cache to be processed by the query functions: getSize, getData, visit
template <typename ID, typename GetID>
void LRUMultiCache<ID, GetID>::setQueryCache(ID id)
{
  CacheMutex::scoped_lock lock(mutex_, /*is_writer=*/true);
  queryCache_ = id;
}

/// Get the size of the cache associated with the specified ID
template <typename ID, typename GetID>
size_t LRUMultiCache<ID, GetID>::getSize() const
{
  unsigned int size;

  { // Lock
  CacheMutex::scoped_lock lock(mutex_, /*is_writer=*/false);

  typename MultiCache::iterator iter = cache_.find(queryCache_);

  if(iter == cache_.end())
  {
    throw NoEntryExistsException("A cache entry for the selected query cache ID does not exist");
  }

  size = (*iter).size();
  } // Unlock

  return size;
}

/// Add a message to the cache
template <typename ID, typename GetID>
void LRUMultiCache<ID, GetID>::addData(const Format* fmt)
{
  CacheMutex::scoped_lock lock(mutex_, /*is_writer=*/true);

  ID id = GetID(fmt);

  FormatCache& formatCache = cache_[id];

  while(formatCache.size() >= maxsize_)
  {
    Format* front = cache_.front();
    cache_.pop_front();
    front->getCreator()->deallocate(front);
  }

  Format* copy = fmt->clone(this);
  cache_.push_back(copy);
}

/// Visit items in the cache
template <typename ID, typename GetID>
void LRUMultiCache<ID, GetID>::visit(FormatVisitor& visitor)
{
  CacheMutex::scoped_lock lock(mutex_, /*is_writer=*/true);

  typename MultiCache::iterator mciter = cache_.find(queryCache_);

  if(mciter == cache_.end())
  {
    throw NoEntryExistsException("A cache entry for the selected query cache ID does not exist");
  }

  for(FormatCache::iterator iter = mciter->second.begin(); iter != mciter->second.end(); ++iter)
  {
    (*iter)->accept(visitor);
  }
}

/// Visit items in the cache
template <typename ID, typename GetID>
void LRUMultiCache<ID, GetID>::visit(ConstFormatVisitor& visitor) const
{
  CacheMutex::scoped_lock lock(mutex_, /*is_writer=*/false);

  typename MultiCache::const_iterator mciter = cache_.find(queryCache_);

  if(mciter == cache_.end())
  {
    throw NoEntryExistsException("A cache entry for the selected query cache ID does not exist");
  }

  for(FormatCache::const_iterator iter = mciter->second.begin(); iter != mciter->second.end(); ++iter)
  {
    (*iter)->accept(visitor);
  }
}
