#include <tbb/parallel_for.h>
#include <tbb/partitioner.h>
#include "Dispatcher.h"
#include "Exceptions.h"
#include "Source.h"
#include "Monitor.h"

Monitor::Monitor()
{
  stopped_ = 0;

#ifndef WIN32
  maxHandle_ = 0;
  FD_ZERO(&readHandles_);
  FD_ZERO(&writeHandles_);
#endif
}

Monitor::~Monitor()
{
}

void Monitor::addSourceMonitor(Source* src, IODirection dir, Delegate* del)
{
  ResourceLock::scoped_lock lock(sourceLock_, /*is_writer=*/true);

  // Add handles
  Source::Handle hnd;
  if(dir == READ)
  {
    hnd = src->getReadHandle();
#ifdef WIN32
    handles_.push_back(hnd);
#else
    FD_SET(hnd, &readHandles_);
#endif
  }
  else
  {
    hnd = src->getWriteHandle();
#ifdef WIN32
    handles_.push_back(hnd);
#else
    FD_SET(hnd, &writeHandles_);
#endif
  }

#ifndef WIN32
  if(hnd > maxHandle_)
  {
    maxHandle_ = hnd;
  }

  sources_.push_back(new SourceRecord(src, dir, del));
#endif
}

void Monitor::removeSourceMonitor(Source* src, IODirection dir)
{
  ResourceLock::scoped_lock lock(sourceLock_, /*is_writer=*/true);

  SourceArray::iterator iter;
  for(iter = sources_.begin(); iter != sources_.end(); ++iter)
  {
    if((*iter)->source_ == src && (*iter)->direction_ == dir)
    {
      break;
    }
  }

  if(iter == sources_.end())
  {
    throw UnrecognizedSourceException("The specified source is not contained by the Monitor");
  }
  else
  {
    sources_.erase(iter);

    delete *iter;

    rebuildHandleContainers();
  }
}

void Monitor::rebuildHandleContainers()
{
  // Rebuild the handle containers
#ifdef WIN32
  handles_.clear();
#else
  FD_ZERO(&readHandles_);
  FD_ZERO(&writeHandles_);
  maxHandle_ = 0;
#endif

  for(SourceArray::iterator iter = sources_.begin(); iter != sources_.end(); ++iter)
  {
    Source::Handle hnd;

    if((*iter)->direction_ == READ)
    {
      hnd = (*iter)->source_->getReadHandle();
#ifdef WIN32
      handles_.push_back(hnd);
#else
      FD_SET(hnd, &readHandles_);
#endif
    }
    else
    {
      hnd = (*iter)->source_->getWriteHandle();
#ifdef WIN32
      handles_.push_back(hnd);
#else
      FD_SET(hnd, &writeHandles_);
#endif
    }

    if(hnd > maxHandle_)
    {
      maxHandle_ = hnd;
    }
  }
}

void Monitor::processRemovals(const SourceRemovals& removals)
{
  bool rebuild = false;

  for(SourceRemovals::const_iterator target = removals.begin(); target != removals.end(); ++target)
  {
    SourceArray::iterator remove = std::find(sources_.begin(), sources_.end(), target->first);
    if(remove != sources_.end())
    {
      sources_.erase(remove);

      rebuild = true;
    }
  }

  if(rebuild)
  {
    rebuildHandleContainers();
  }
}

void Monitor::SourceArrayForOp::operator()(const tbb::blocked_range<size_t>& r) const
{
  for(size_t i = r.begin(); i != r.end(); ++i)
  {
#ifdef WIN32
    if(sources_[i]->direction_ == READ && WaitForSingleObject(source_[i]->source_->getReadHandle(), 0))
#else
    if(sources_[i]->direction_ == READ && FD_ISSET(sources_[i]->source_->getReadHandle(), readHandles_))
#endif
    {
      try
      {
        Format* fmt = sources_[i]->source_->read();
        if(fmt != NULL)
        {
          sources_[i]->delegate_->dispatch(fmt);
        }
      }
      catch(EOFException eof)
      {
        eofremovals_.push_back(std::make_pair(sources_[i], eof.getMessage()));
      }
      catch(ReadFailedException err)
      {
        errorremovals_.push_back(make_pair(sources_[i], err.getMessage()));
      }
    }
#ifdef WIN32
    else if(sources_[i]->direction_ == WRITE && WaitForSingleObject(sources_[i]->source_->getWriteHandle(), 0))
#else
    else if(sources_[i]->direction_ == WRITE && FD_ISSET(sources_[i]->source_->getWriteHandle(), writeHandles_))
#endif
    {
      try
      {
        // Write message from queue
        sources_[i]->source_->writeQueued();
      }
      catch(EOFException eof)
      {
        eofremovals_.push_back(make_pair(sources_[i], eof.getMessage()));
      }
      catch(WriteFailedException err)
      {
        errorremovals_.push_back(make_pair(sources_[i], err.getMessage()));
      }
    }
  }
}

void Monitor::run()
{
  ResourceLock::scoped_lock lock;

#ifndef WIN32
  struct timeval timeout = { 1, 0 };
#endif

  stopped_ = 0;

  while(!stopped_)
  {
    lock.acquire(sourceLock_, /*is_writer*/false);

#ifdef WIN32
    int result = WaitForMultipleObjects(handles_.size(), &handles_[0], FALSE, 1000);

    if(result == WAIT_FAILED)
    {
      throw MonitorFailedException("Monitor failed to retrieve source state");
    }
    else if(result != WAIT_TIMEOUT)
    {
      tbb::parallel_for(tbb::blocked_range<size_t>(0, sources_.size()), SourceArrayForOp(sources_, eofremovals_, errorremovals_), tbb::auto_partitioner());
    }
#else
    int ndfs = maxHandle_ + 1;
    fd_set readHandles = readHandles_;
    fd_set writeHandles = writeHandles_;

    int result = select(ndfs, &readHandles, &writeHandles, NULL, &timeout);

    if(result == -1)
    {
      throw MonitorFailedException("Monitor failed to retrieve source state");
    }
    else if(result > 0)
    {
      tbb::parallel_for(tbb::blocked_range<size_t>(0, sources_.size()), SourceArrayForOp(sources_, eofremovals_, errorremovals_, &readHandles, &writeHandles), tbb::auto_partitioner());
    }
#endif

    // Check for sources to remove
    processRemovals(eofremovals_);
    eofsources_.insert(eofsources_.end(), eofremovals_.begin(), eofremovals_.end());
    eofremovals_.clear();

    processRemovals(errorremovals_);
    errorsources_.insert(errorsources_.end(), errorremovals_.begin(), errorremovals_.end());
    errorremovals_.clear();

    lock.release();
  }
}
