#ifndef MONITOR_H
#define MONITOR_H

#ifdef WIN32
#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif
#include <windows.h>
#else
#include <sys/select.h>
#endif

#include <vector>
#include <tbb/atomic.h>
#include <tbb/blocked_range.h>
#include <tbb/spin_rw_mutex.h>

class Delegate;

class Monitor
{
public:
  enum IODirection
  {
    READ = 0,
    WRITE = 1
  };

protected:
  // Information for a monitored source
  class SourceRecord
  {
  public:
    Source* source_;
    IODirection direction_;
    Delegate* delegate_;
  public:
    SourceRecord(Source* src, IODirection dir, Delegate* del) : source_(src), direction_(dir), delegate_(del) { }
  };

  // Aray of source records to track currently monitored sources
  typedef std::vector< SourceRecord* > SourceArray;

  // List for storing sources which have either reached EOF or have raised errors and must be removed from monitoring at end of run loop
  typedef std::vector< std::pair< SourceRecord*, std::string > > SourceRemovals;

protected:
  // Operation for processing sources with parallel_for
  class SourceArrayForOp
  {
  protected:
    SourceArray& sources_;
    SourceRemovals& eofremovals_;
    SourceRemovals& errorremovals_;
#ifndef WIN32
    fd_set* readHandles_;
    fd_set* writeHandles_;
#endif
  public:
#ifdef WIN32
    SourceArrayForOp(SourceArray& sources, SourceRemovals& eofremovals, SourceRemovals& errorremovals) : sources_(sources), eofremovals_(eofremovals), errorremovals_(errorremovals) { }
#else
    SourceArrayForOp(SourceArray& sources, SourceRemovals& eofremovals, SourceRemovals& errorremovals, fd_set* readHandles, fd_set* writeHandles) : sources_(sources), eofremovals_(eofremovals), errorremovals_(errorremovals), readHandles_(readHandles), writeHandles_(writeHandles) { }
#endif

    void operator()(const tbb::blocked_range<size_t>& r) const;
  };

  // Removes the specified sources from the source array
  void processRemovals(const SourceRemovals& removals);

  // Rebuilds the lists of read and write handles after a removal; shared functionality for processRemovals and removeSourceMonitor
  void rebuildHandleContainers();

public:
  Monitor();

  ~Monitor();

  void addSourceMonitor(Source* src, IODirection dir, Delegate* del);

  void removeSourceMonitor(Source* src, IODirection dir);

  size_t getNumMonitoredSources() const { return sources_.size(); }

  size_t getNumEOFStates() const { return eofsources_.size(); }

  size_t getNumErrorStates() const { return eofsources_.size(); }

  void run();

  void stop() { stopped_ = 1; }

private:
  typedef tbb::spin_rw_mutex ResourceLock;

  ResourceLock sourceLock_;

  mutable ResourceLock dispatcherLock_;

  tbb::atomic<int> stopped_;

  SourceArray sources_;

  // Vector for tracking sources in the EOF state
  SourceRemovals eofsources_;

  // Vector for tracking sources in the error state
  SourceRemovals errorsources_;

  // Vector for receiving EOF state sources from the parallel_for operation class
  SourceRemovals eofremovals_;

  // Vector for receiving error state sources from the parallel_for operation class
  SourceRemovals errorremovals_;

#ifdef WIN32
  std::vector< HANDLE > handles_;
#else
  int maxHandle_;

  fd_set readHandles_;

  fd_set writeHandles_;
#endif
};

#endif // MONITOR_H
