#include "PipelineFunctions.h"
#include "Pipeline.h"

void* Pipeline::Initiator::operator()(void*)
{
  // Stop processing pending data if task is suspended
  if(pipeline_->suspended_)
  {
    return NULL;
  }

  Format* fmt;
  return pipeline_->data_.pop_if_present(fmt) ? fmt : NULL;
}

Pipeline::Pipeline()
: init_(this),
  dirty_(false),
  term_(NULL)
{
  suspended_ = FALSE;
  resumed_ = FALSE;
  running_ = FALSE;
  signalLock_.acquire(signal_);

  pipeline_ = new tbb::pipeline;
  pipeline_->add_filter(init_);
}

Pipeline::Pipeline(Terminator* term)
: init_(this),
  dirty_(false),
  term_(term)
{
  suspended_ = FALSE;
  resumed_ = FALSE;
  running_ = FALSE;
  signalLock_.acquire(signal_);

  pipeline_ = new tbb::pipeline;
  pipeline_->add_filter(init_);
  pipeline_->add_filter(*term_);
}

Pipeline::~Pipeline()
{
  delete pipeline_;
}

void Pipeline::getTerminator(Terminator*& term) const
{
  FunctionListLock::scoped_lock lock(funcsMutex_, /*is_writer=*/false);
  term = term_;
}

// Locked with function lock to prevent change while pipeline is being constructed
void Pipeline::setTerminator(Terminator* term)
{
  FunctionListLock::scoped_lock lock(funcsMutex_, /*is_writer=*/true);
  term_ = term;
  dirty_ = true;
}

unsigned int Pipeline::getFunctionListSize() const
{
  unsigned int size;

  { // Lock
  FunctionListLock::scoped_lock lock(funcsMutex_, /*is_writer=*/false);
  size = funcs_.size();
  } // Unlock

  return size;
}

// Visitor to lock and process all elemnts of list safely; safe alternative to providing iterators to external callers
void Pipeline::visitFunctions(FunctionVisitor& visitor)
{
  FunctionListLock::scoped_lock lock(funcsMutex_, /*is_writer=*/true);
  for(FunctionList::iterator iter = funcs_.begin(); iter != funcs_.end(); ++iter)
  {
    visitor(*iter);
  }
}

void Pipeline::visitFunctions(const ConstFunctionVisitor& visitor) const
{
  FunctionListLock::scoped_lock lock(funcsMutex_, /*is_writer=*/false);
  for(FunctionList::const_iterator iter = funcs_.begin(); iter != funcs_.end(); ++iter)
  {
    visitor(*iter);
  }
}

void Pipeline::getFunction(unsigned int pos, Function*& func)
{
  FunctionListLock::scoped_lock lock(funcsMutex_, /*is_writer=*/false);
  FunctionList::iterator iter = funcs_.begin();
  advance(iter, pos);
  func = *iter;
}

void Pipeline::addFunction(Function* func)
{
  FunctionListLock::scoped_lock lock(funcsMutex_, /*is_writer=*/true);
  funcs_.push_back(func);
  dirty_ = true;
}

void Pipeline::insertFunction(unsigned int pos, Function* func)
{
  FunctionListLock::scoped_lock lock(funcsMutex_, /*is_writer=*/true);
  FunctionList::iterator iter = funcs_.begin();
  advance(iter, pos);
  funcs_.insert(iter, func);
  dirty_ = true;
}

void Pipeline::removeFunction(unsigned int pos)
{
  FunctionListLock::scoped_lock lock(funcsMutex_, /*is_writer=*/true);
  FunctionList::iterator iter = funcs_.begin();
  advance(iter, pos);
  funcs_.erase(iter);
  dirty_ = true;
}

void Pipeline::moveFunction(unsigned int fromPos, unsigned int toPos)
{
  FunctionListLock::scoped_lock lock(funcsMutex_, /*is_writer=*/true);

  FunctionList::iterator iter = funcs_.begin();
  advance(iter, fromPos);

  Function* func = *iter;
  funcs_.erase(iter);

  iter = funcs_.begin();
  advance(iter, toPos);

  funcs_.insert(iter, func);

  dirty_ = true;
}

void Pipeline::clearFunctions()
{
  FunctionListLock::scoped_lock lock(funcsMutex_, /*is_writer=*/true);
  funcs_.clear();
  dirty_ = true;
}

bool Pipeline::suspend()
{
  return suspended_.compare_and_swap(TRUE, FALSE) == FALSE;
}

bool Pipeline::resume()
{
  if(suspended_ == TRUE)
  {
    if(resumed_.compare_and_swap(TRUE, FALSE) == FALSE)
    {
      //SIGNALRUNTHREAD
      signalLock_.release();
    }
  }

  return false;
}

void Pipeline::commitFunctionChanges()
{
//  pipeline_->clear(); //This did not work, pipeline is in bad state and must be recreated
  delete pipeline_;
  pipeline_ = new tbb::pipeline;
  pipeline_->add_filter(init_);

  for(FunctionList::iterator iter = funcs_.begin(); iter != funcs_.end(); ++iter)
  {
    pipeline_->add_filter(*(*iter));
  }

  if(term_ != NULL)
  {
    pipeline_->add_filter(*term_);
  }
}

void Pipeline::run(size_t limit, size_t max_num_live_pipelines)
{
  { // Lock
  FunctionListLock::scoped_lock lock(funcsMutex_, /*is_writer=*/false);
  if(dirty_)
  {
    commitFunctionChanges();
    dirty_ = false;
  }
  } // Unlock

  running_ = TRUE;

  for(;;)
  {
    init_.setLimit(limit);
    pipeline_->run(max_num_live_pipelines);

    if(suspended_ != TRUE)
    {
      // If data was added while pipeline was active, process that data
      if(data_.empty())
      {
        running_ = FALSE;
        break;
      }
    }
    else
    {
      //WAITFORRESUMESIGNAL
      running_ = FALSE;
      signalLock_.acquire(signal_);

      { // Lock
      FunctionListLock::scoped_lock lock(funcsMutex_, /*is_writer=*/false);
      if(dirty_)
      {
        commitFunctionChanges();
        dirty_ = false;
      }
      } // Unlock

      running_ = TRUE;
      resumed_ = FALSE;
      suspended_ = FALSE;
    }
  }
}
