#ifndef PIPELINE_H
#define PIPELINE_H

#include <list>
#include <tbb/atomic.h>
#include <tbb/concurrent_queue.h>
#include <tbb/pipeline.h>
#include <tbb/spin_rw_mutex.h>
#include <tbb/mutex.h>

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

class Format;
class Function;
class FunctionVisitor;
class ConstFunctionVisitor;
class Terminator;

/* Only one task may interact with the pipeline; options for pipeline reconfiguration design are:
   Option 1
    1.  Invoke "begin" reconfiguration: this causes pipeline to stop issuing any new items
        and complete any items that are currently in-flight; pipeline then waits for configuration "end"
    2.  Modifications are made to pipeline through the API
    3.  Invoke "end" reconfiguration: causes pipeline to execute reconfiguration and then resume

    Option 2
    1.  Modifications are made to pipeline through API; each modification marks pipeline as "dirty"
    2.  Invoke "reconfigure" command which halts pipeline and resumes pipeline; recofnigures automatically
        at pipeline restart pipeline
    3.  Can optionally add an argument to modification function to execute "reconfigure" automatically

    Option 3
    1.  Combination of 1 and 2
    2.  Optionally call "suspend" to pause pipeline while reconfiguration commences
    3.  Make modifications; modifications mark pipeline as dirty
    4.  if "suspended" then "resume" to start pipeline else "reconfigure" which is an alias to
        "suspend; resume"; reconfiguration is automatic in "run" if pipeline is dirt

    Potential proplems with multiple tasks calling resume/suspend
      - Require thread that calls suspend to be thread that calls resume
      - Threads attemting to suspend while other threads have suspended, must wait
        for current thread to resume; Only one suspender at a time
      - Reconfiguration becomes a serialized process

    Option 2 has problem that multiple tasks can also simultaneously modify pipeline
      - A-B-A problem
      - Making modification operations atomic is not enough
      - Serialization of process means no two threads can modify state at once; no unexpected
        modifications by a separate thread

    Option 1 has problem that one thread can lock pipeline and never unlock pipeline
      - Deadlock
      - Instead, do not lock
      - This non-locking model allows modifications to be made while items are in-flight, and commited
        with minimal suspension time; or to suspend while changes are made to minimize number of
        items processed with "outdated" pipeline

    Option 3 has problem that optional locking is complicated
      - Need a way to stop thread who does not have lock from modifying pipeline; possibly
        requires two variants of function call with args for identifying thread who wants access
      - Instead, do not lock
      - This non-locking model allows modifications to be made while items are in-flight, and commited
        with minimal suspension time; or to suspend while changes are made to minimize number of
        items processed with "outdated" pipeline

    Solution generalize Option 3
    1.  Decouple locking from suspension
    2.  Interpreter handles locking and handles deadlock case
      a.  User can see who has lock
      b.  User can request lock release
        i.   Other user gets notification with required y/n response within n seconds
        ii.  Failure to respond automatically releases lock
      c.  User can close remote connection
        i.   All locks are released on cleanup
    3.  Interpreter required to handle serialization
      a.  Interpreter is single thread executing pipeline changes
      b.  Handle race conditions a interpreter level
    4.  Changes to pipeline mark as "dirty"
    5.  Pipeline "run" reconstructs "dirty" pipeline before starting
      a.  suspend/resume trigger the restart
    6.  Modifier/access/commit functions need some (rw)? mutex protection for run thread vs modifier thread safety
*/

class Pipeline
{
public:
  typedef std::list< Function* > FunctionList;

  typedef tbb::concurrent_queue< Format* > DataQueue;

  Pipeline();

  Pipeline(Terminator* term);

  ~Pipeline();

  unsigned int getDataSize() const { return data_.size(); }

  void addData(Format* fmt) { data_.push(fmt); }

  void getTerminator(Terminator*& term) const;

  void setTerminator(Terminator* term);

  unsigned int getFunctionListSize() const;

  // Visitor to write lock and process all elemnts of list safely
  // Safe alternative to providing iterators to external callers
  // Not a paraller loop
  void visitFunctions(FunctionVisitor& visitor);

  // Visitor to read lock and process all elemnts of list safely
  // Safe alternative to providing iterators to external callers
  // Not a paraller loop
  void visitFunctions(const ConstFunctionVisitor& visitor) const;

  void getFunction(unsigned int pos, Function*& func);

  void addFunction(Function* func);

  void insertFunction(unsigned int pos, Function* func);

  void removeFunction(unsigned int pos);

  void moveFunction(unsigned int frompos, unsigned int topos);

  void clearFunctions();

  bool suspend();

  // Could block until processing has resumed if necessary
  bool resume();

  // The only function to access the tbb::pipeline
  void run(size_t limit, size_t max_num_live_pipelines);

  bool isSuspended() const { return suspended_ == TRUE; }

  bool isRunning() const { return running_ == TRUE; }
protected:
  class Initiator : public tbb::filter
  {
  protected:
    size_t limit_;          // The limit should only be changed in the Pipeline::run function before the pipeline starts
    Pipeline* pipeline_;

  public:
    Initiator(Pipeline* pipeline) : tbb::filter(/*is_serial*/true), limit_(0), pipeline_(pipeline) { }

    size_t getLimit() const { return limit_; }

    void setLimit(size_t limit) { limit_ = limit; }

    virtual void* operator()(void*);
  };

  tbb::pipeline* pipeline_;

  DataQueue data_;

  Initiator init_;
private:
  typedef tbb::spin_rw_mutex FunctionListLock;

  typedef tbb::mutex ResumeSignal;

  // Only modified within the function list lock
  bool dirty_;

  // Indicates that pipeline is suspended, only enabled by suspend() and disabled by run()
  tbb::atomic<int> suspended_;

  // Indicates that pipeline is already resumed, only enabled by resume() and disabled by run()
  // Prevents resmume from executing more than once
  tbb::atomic<int> resumed_;

  // Indicates that pipeline is running, only modified by run()
  tbb::atomic<int> running_;

  FunctionList funcs_;

  mutable FunctionListLock funcsMutex_;

  ResumeSignal signal_;

  ResumeSignal::scoped_lock signalLock_;

  Terminator* term_;

  void commitFunctionChanges();
};

#endif
