#ifndef PIPELINEDISPATCHER_H
#define PIPELINEDISPATCHER_H

#include <vector>
#include <tbb/atomic.h>
#include <tbb/blocked_range.h>
#include <tbb/concurrent_hash_map.h>
#include "Creator.h"
#include "Dispatcher.h"
#include "Exceptions.h"
#include "Pipeline.h"
#include "Format.h"

class Source;

typedef unsigned int FormatType;

struct FormatTypeHashCompare
{
  static size_t hash(FormatType x) { return x; }

  static bool equal(const FormatType x, const FormatType y ) { return x == y; }
};

typedef std::vector<Pipeline*> Pipelines;
typedef tbb::concurrent_hash_map<FormatType, Pipeline*, FormatTypeHashCompare> PipelineGroup;
typedef tbb::concurrent_hash_map<FormatType, Pipelines, FormatTypeHashCompare> PipelinesGroup;

// Delegate for one pipeline per source
class PipelineDelegate : public Delegate
{
protected:
  Pipeline* pipeline_;

public:
  PipelineDelegate(Pipeline* pipeline) : pipeline_(pipeline) { }

  Pipeline* getPipeline() const { return pipeline_; }

  virtual void dispatch(Format* fmt)
  {
    if(pipeline_ != NULL)
    {
      pipeline_->addData(fmt);
    }
    else
    {
      throw UnregisteredFormatException("The specified format has not been registered with the delegate");
    }
  }

  virtual void run(size_t limit, size_t max_num_live_pipelines)
  {
    if(pipeline_->getDataSize() > 0)
    {
      pipeline_->run(limit, max_num_live_pipelines);
    }
  }
};

// Delegate for one pipeline per format; n formats per source
class PipelineGroupDelegate : public Delegate
{
protected:
  PipelineGroup* group_;

public:
  PipelineGroupDelegate(PipelineGroup* group) : group_(group) { }

  PipelineGroup* getPipelineGroup() const { return group_; }

  virtual void dispatch(Format* fmt)
  {
    PipelineGroup::accessor acc;
    if(group_->find(acc, fmt->getType()))
    {
      acc->second->addData(fmt);
      acc.release();
    }
    else
    {
      throw UnregisteredFormatException("The specified format has not been registered with the delegate");
    }
  }

  virtual void run(size_t limit, size_t max_num_live_pipelines);
};

// Delegate for n pipelines per source
// Pipelines may not be safely modified when delegate is registered with an active monitor/dispatcher
class PipelinesDelegate : public Delegate, public Creator
{
protected:
  Pipelines* pipelines_;

public:
  PipelinesDelegate(Pipelines* pipelines) : pipelines_(pipelines) { }

  Pipelines* getPipelines() const { return pipelines_; }

  virtual void dispatch(Format* fmt)
  {
    if(pipelines_ != NULL)
    {
      for(Pipelines::iterator iter = pipelines_->begin(); iter != pipelines_->end(); ++iter)
      {
        (*iter)->addData(fmt->clone(this));
      }

      fmt->getCreator()->deallocate(fmt);
    }
    else
    {
      fmt->getCreator()->deallocate(fmt);
      throw UnregisteredFormatException("The specified format has not been registered with the delegate");
    }
  }

  virtual void run(size_t limit, size_t max_num_live_pipelines);
};

// Delegate for n pipelines per format; n formats per source
// Pipelines may be safely modified when accessed through the accessor and const_accessor objects
class PipelinesGroupDelegate : public Delegate, public Creator
{
protected:
  PipelinesGroup* group_;

public:
  PipelinesGroupDelegate(PipelinesGroup* group) : group_(group) { }

  PipelinesGroup* getPipelinesGroup() const { return group_; }

  virtual void dispatch(Format* fmt)
  {
    PipelinesGroup::accessor acc;
    if(group_->find(acc, fmt->getType()))
    {
      Pipelines& pipelines = acc->second;
      for(Pipelines::iterator iter = pipelines.begin(); iter != pipelines.end(); ++iter)
      {
        (*iter)->addData(fmt->clone(this));
      }

      acc.release();
      fmt->getCreator()->deallocate(fmt);
    }
    else
    {
      fmt->getCreator()->deallocate(fmt);
      throw UnregisteredFormatException("The specified format has not been registered with the delegate");
    }
  }

  virtual void run(size_t limit, size_t max_num_live_pipelines);
};

class PipelineDispatcher
{
public:
  PipelineDispatcher() { running_ = 0; }

  void addSourceDelegate(Source* src, Delegate* del)
  {
    Delegates::accessor acc;

    if(!delegates_.insert(acc, src))
    {
      throw EntryExistsException("A delegate entry for the specified source already exists");
    }

    acc->second = del;
  }

  void removeSourceDelegate(Source* src)
  {
    if(!delegates_.erase(src))
    {
      throw NoEntryExistsException("A delegate entry for the specified source does not exist");
    }
  }

  void dispatch(Source* src, Format* fmt)
  {
    Delegates::accessor acc;
    if(delegates_.find(acc, src))
    {
      acc->second->dispatch(fmt);
      acc.release();
    }
    else
    {
      throw UnregisteredFormatException("The specified format has not been registered with the dispatcher");
    }
  }

  void run();

  void stop() { running_ = 0; }

protected:
  struct SourceHashCompare
  {
    static size_t hash(Source* x) { return reinterpret_cast<size_t>(x); }

    static bool equal(const Source* x, const Source* y ) { return x == y; }
  };

  typedef tbb::concurrent_hash_map<Source*, Delegate*, SourceHashCompare> Delegates;

  class PipelineDispatcherForOp
  {
  public:
    PipelineDispatcherForOp() { }

    void operator()(PipelineDispatcher::Delegates::range_type& r) const;
  };

protected:
  Delegates delegates_;

  tbb::atomic<int> running_;
};

#endif // PIPELINEMANAGER_H
