#ifndef PIPELINEFUNCTIONS_H
#define PIPELINEFUNCTIONS_H

// For filter group
#include <vector>
#include <tbb/spin_rw_mutex.h>
#include <tbb/parallel_reduce.h>
#include <tbb/blocked_range.h>

#include <tbb/pipeline.h>
#include "Creator.h"

class Format;

class Function : public tbb::filter
{
  public:
    Function(bool is_serial) : tbb::filter(is_serial) { }
    virtual void* operator()(void* item) = 0;
};

class Filter : public Function
{
public:
  // Constructor - set filter mode as serial or parallel
  Filter(bool is_serial) : Function(is_serial) { }

  // Filter based on format contents
  virtual bool operator()(const Format*) = 0;
private:
  virtual void* operator()(void* item)
  {
    if(item == NULL)
    {
      return NULL;
    }

    Format* format = static_cast< Format* >(item);
    if((*this)(static_cast< Format* >(format)))
    {
      return item;
    }
    else
    {
      format->getCreator()->deallocate(format);
      return NULL;
    }
  }
};

class Transform : public Function
{
public:
  // Constructor - set filter mode as serial or parallel
  Transform(bool is_serial) : Function(is_serial) { }

  // Transform format contents
  virtual bool operator()(Format*) = 0;
private:
  virtual void* operator()(void* item)
  {
    if(item == NULL)
    {
      return NULL;
    }

    Format* format = static_cast< Format* >(item);
    if((*this)(static_cast< Format* >(format)))
    {
      return item;
    }
    else
    {
      format->getCreator()->deallocate(format);
      return NULL;
    }
  }
};

class Translator : public Function, public Creator
{
public:
  // Constructor - filter mode is serial
  Translator(bool is_serial) : Function(is_serial) { }

  // Convert to a new format
  virtual Format* operator()(const Format*) = 0;
private:
  virtual void* operator()(void* item)
  {
    if(item == NULL)
    {
      return NULL;
    }

    Format* format = static_cast< Format* >(item);
    Format* result = (*this)(format);
    format->getCreator()->deallocate(format);
    return result;
  }
};

class Terminator : public tbb::filter, public Creator
{
public:
  // Constructor - filter mode is serial
  Terminator(bool is_serial) : tbb::filter(is_serial) { }

  virtual bool operator()(Format*) = 0;
private:
  // The terminator does not implicitly deallocate the data because it could have transferred ownership to another object
  virtual void* operator()(void* item)
  {
    if(item == NULL)
    {
      return NULL;
    }

    return (*this)(static_cast< Format* >(item)) ? item : NULL;
  }
};

class FunctionVisitor
{
public:
  virtual ~FunctionVisitor() { }

  virtual void operator()(Function*) = 0;
};

class ConstFunctionVisitor
{
public:
  virtual ~ConstFunctionVisitor() { }

  virtual void operator()(const Function*) const = 0;
};

class FilterGroup : public Filter
{
protected:
  typedef std::vector<Filter*> Filters;

  typedef tbb::spin_rw_mutex FiltersLock;

  class FilterGroupOp
  {
  protected:
    bool result_;

    const Format* format_;

    Filters filters_;

  public:
    FilterGroupOp(const Format* format, Filters& filters) : result_(true), format_(format), filters_(filters) { }

    FilterGroupOp(FilterGroupOp& f, tbb::split) : result_(true), format_(f.format_), filters_(f.filters_) { }

    bool getResult() const { return result_; }

    void operator()(const tbb::blocked_range<size_t>& r)
    {
      for(size_t i = r.begin(); i != r.end(); ++i)
      {
        if(!(*filters_[i])(format_))
        {
          result_ = false;
        }
      }
    }

    void join(FilterGroupOp& rhs)
    {
      if(!rhs.result_)
      {
        result_ = false;
      }
    }
  };

protected:
  Filters filters_;

  mutable FiltersLock filtersMutex_;

public:
  // Constructor - set filter mode as serial or parallel
  FilterGroup(bool is_serial) : Filter(is_serial) { }

  size_t getFilterListSize() const
  {
    size_t size;

    { // Lock
    FiltersLock::scoped_lock lock(filtersMutex_, /*is_writer=*/false);
    size = filters_.size();
    } // Unlock

    return size;
  }

  // Visitor to lock and process all elemnts of list safely; safe alternative to providing iterators to external callers
  void visitFilters(FunctionVisitor& visitor)
  {
    FiltersLock::scoped_lock lock(filtersMutex_, /*is_writer=*/true);
    for(Filters::iterator iter = filters_.begin(); iter != filters_.end(); ++iter)
    {
      visitor(*iter);
    }
  }

  void visitFilters(const ConstFunctionVisitor& visitor) const
  {
    FiltersLock::scoped_lock lock(filtersMutex_, /*is_writer=*/false);
    for(Filters::const_iterator iter = filters_.begin(); iter != filters_.end(); ++iter)
    {
      visitor(*iter);
    }
  }

  void getFilter(size_t pos, Filter*& filter)
  {
    FiltersLock::scoped_lock lock(filtersMutex_, /*is_writer=*/false);
    Filters::iterator iter = filters_.begin();
    advance(iter, pos);
    filter = *iter;
  }

  void addFilter(Filter* filter)
  {
    FiltersLock::scoped_lock lock(filtersMutex_, /*is_writer=*/true);
    filters_.push_back(filter);
  }

  void insertFilter(size_t pos, Filter* filter)
  {
    FiltersLock::scoped_lock lock(filtersMutex_, /*is_writer=*/true);
    Filters::iterator iter = filters_.begin();
    advance(iter, pos);
    filters_.insert(iter, filter);
  }

  void removeFilter(size_t pos)
  {
    FiltersLock::scoped_lock lock(filtersMutex_, /*is_writer=*/true);
    Filters::iterator iter = filters_.begin();
    advance(iter, pos);
    filters_.erase(iter);
  }

  void moveFilter(size_t fromPos, size_t toPos)
  {
    FiltersLock::scoped_lock lock(filtersMutex_, /*is_writer=*/true);

    Filters::iterator iter = filters_.begin();
    advance(iter, fromPos);

    Filter* filter = *iter;
    filters_.erase(iter);

    iter = filters_.begin();
    advance(iter, toPos);

    filters_.insert(iter, filter);
  }

  void clearFilters()
  {
    FiltersLock::scoped_lock lock(filtersMutex_, /*is_writer=*/true);
    filters_.clear();
  }

  virtual bool operator()(const Format* format)
  {
    FiltersLock::scoped_lock lock(filtersMutex_, /*is_writer=*/false);
    FilterGroupOp op(format, filters_);
    tbb::parallel_reduce(tbb::blocked_range<size_t>(0, filters_.size()), op, tbb::auto_partitioner());
    return op.getResult();
  }
};

#endif // PIPELINEFUNCTIONS_H
