#ifndef POOLALLOCATOR_H
#define POOLALLOCATOR_H

#include <cassert>
#include <tbb/concurrent_queue.h>

#ifndef NOT_USING_SCALABLE_ALLOCATOR
#include <tbb/scalable_allocator.h>
#endif

template <typename T>
class PoolAllocator
{
protected:
  typedef tbb::concurrent_queue<T*> MemoryPool;

  MemoryPool pool_;

  int maxsize_;

public:
  PoolAllocator(int maxsize) : maxsize_(maxsize) { }

  ~PoolAllocator() { clear(); }

  int getMaxSize() const { return maxsize_; }

  void* allocate(size_t size)
  {
    T* val;
#ifdef NOT_USING_SCALABLE_ALLOCATOR
    return pool_.pop_if_present(val) ? val : ::new char [size];
#else
    return pool_.pop_if_present(val) ? val : scalable_malloc(size);
#endif
  }

  void deallocate(void* val)
  {
    if(pool_.size() < maxsize_)
    {
      T* value = static_cast<T*>(val);
      pool_.push(value);
    }
    else
    {
#ifdef NOT_USING_SCALABLE_ALLOCATOR
      ::delete [] reinterpret_cast<char*>(val);
#else
      scalable_free(val);
#endif
    }
  }

  void clear()
  {
    T* val;
    while(!pool_.empty())
    {
      pool_.pop(val);
#ifdef NOT_USING_SCALABLE_ALLOCATOR
      ::delete [] reinterpret_cast<char*>(val);
#else
      scalable_free(val);
#endif
    }
  }
};

// Macro to create a pool allocator for
#define POOL_DECLARE(NAME) \
static PoolAllocator<NAME> __poolallocator; \
void* operator new(size_t size) { assert(size == sizeof(NAME)); return NAME::__poolallocator.allocate(size); } \
void operator delete(void* val, size_t size) { assert(size == sizeof(NAME)); NAME::__poolallocator.deallocate(val); }

// Macro to define a pool allocator
#define POOL_DEFINE(NAME, SIZE) \
PoolAllocator<NAME> NAME::__poolallocator(SIZE);

#endif
