#include "Cache.h"
#include "Classifier.h"
#include "Format.h"
#include "Source.h"

Source::Handle Source::getReadHandle() const
{
  if(mode_ == OUT)
  {
    throw UnsupportedModeException("Requesting read handle from a write only source");
  }

  return readHandle_;
}

Source::Handle Source::getWriteHandle() const
{
  if(mode_ == IN)
  {
    throw UnsupportedModeException("Requesting write handle from a read only source");
  }

  return writeHandle_;
}

void Source::setReadCache(Cache* cache)
{
  if(mode_ == OUT)
  {
    throw UnsupportedModeException("Specifying read cache for a write only source");
  }

  readCache_ = cache;
}

Cache* Source::getReadCache() const
{
  if(mode_ == OUT)
  {
    throw UnsupportedModeException("Requesting read cache for a write only source");
  }

  return readCache_;
}

void Source::setWriteCache(Cache* cache)
{
  if(mode_ == IN)
  {
    throw UnsupportedModeException("Specifying write cache for a read only source");
  }

  writeCache_ = cache;
}

Cache* Source::getWriteCache() const
{
  if(mode_ == IN)
  {
    throw UnsupportedModeException("Requesting write cache for a read only source");
  }

  return writeCache_;
}

Format* Source::read()
{
  if(mode_ == OUT)
  {
    throw UnsupportedModeException("Attempting to read from a write only source");
  }

  Format* fmt = readRaw();

  if(classifier_)
  {
    Format* tmp = classifier_->fromRaw(fmt);
    fmt->getCreator()->deallocate(fmt);
    fmt = tmp;
  }

  // The classifier could produce nothing
  if(fmt != NULL)
  {
    if(readCache_ != NULL)
    {
      readCache_->addData(fmt);
    }

    fmt->setSource(this);
  }

  return fmt;
}

// Don't delete the format data
// This is not safe to call when the source is
// part of a monitor/pipeline pair.  It exists for use to write a header or
// footer to the file before or after submitting for monitoring.
void Source::write(const Format* fmt)
{
  if(mode_ == IN)
  {
    throw UnsupportedModeException("Attempting to write to a read only source");
  }

  if(writeCache_ != NULL)
  {
    writeCache_->addData(fmt);
  }

  if(classifier_)
  {
    Format* tmp = classifier_->toRaw(fmt);

    // The classifier could produce nothing
    if(tmp != NULL)
    {
      writeRaw(tmp);
      tmp->getCreator()->deallocate(tmp);
    }
  }
  else
  {
    writeRaw(fmt);
  }
}

// Delete the queued format when done - if other threads reference this value, it should be inserted as a copy
void Source::writeQueued()
{
  if(mode_ == IN)
  {
    throw UnsupportedModeException("Attempting to write to a read only source");
  }

  Format* fmt;
  if(writeQueue_.pop_if_present(fmt))
  {
    if(writeCache_ != NULL)
    {
      writeCache_->addData(fmt);
    }

    if(classifier_)
    {
      Format* tmp = classifier_->toRaw(fmt);
      fmt->getCreator()->deallocate(fmt);
      fmt = tmp;
    }

    // The classifier could produce nothing
    if(fmt != NULL)
    {
      writeRaw(fmt);
      fmt->getCreator()->deallocate(fmt);
    }
  }
}

// Delete the queued format when done - if other threads reference this value, it should be inserted as a copy
void Source::enqueueForWrite(Format* fmt)
{
  if(mode_ == IN)
  {
    throw UnsupportedModeException("Attempting to write to a read only source");
  }

  writeQueue_.push(fmt);
}
