#include "Exceptions.h"
#include "RawDatagramFormat.h"
#include "UnicastSource.h"

// OUT Constructor
UnicastSource::UnicastSource(Mode mode)
: Source(mode),
  port_(EPHEMERAL_PORT),
  attached_(false)
{
  if(mode != OUT)
  {
    throw UnsupportedModeException("The specified i/o mode is not valid for the current constructor");
  }

  create();
  state_ = ACTIVE;
}

// OUT Constructor
UnicastSource::UnicastSource(Mode mode, const std::string& addr, unsigned short port)
: Source(mode),
  addr_(addr),
  port_(port),
  attached_(true)
{
  if(mode != OUT)
  {
    throw UnsupportedModeException("The specified i/o mode is not valid for the current constructor");
  }

  create();
  state_ = ACTIVE;
}

// IN or INOUT Constructor
UnicastSource::UnicastSource(Mode mode, unsigned short port)
: Source(mode),
  port_(EPHEMERAL_PORT),
  attached_(false)
{
  if(mode == OUT)
  {
    throw UnsupportedModeException("The specified i/o mode is not valid for the current constructor");
  }

  create();
  bind(IPEndPoint("", port));
  state_ = ACTIVE;
}

// IN or INOUT Constructor
UnicastSource::UnicastSource(Mode mode, const IPEndPoint& endpoint)
: Source(mode),
  port_(EPHEMERAL_PORT),
  attached_(false)
{
  if(mode == OUT)
  {
    throw UnsupportedModeException("The specified i/o mode is not valid for the current constructor");
  }

  create();
  bind(endpoint);
  state_ = ACTIVE;
}

// INOUT Constructor
UnicastSource::UnicastSource(Mode mode, unsigned short localport, const std::string& addr, unsigned short port)
: Source(mode),
  addr_(addr),
  port_(port),
  attached_(true)
{
  if(mode != INOUT)
  {
    throw UnsupportedModeException("The specified i/o mode is not valid for the current constructor");
  }

  create();
  bind(IPEndPoint("", port));
  state_ = ACTIVE;
}

// INOUT Constructor
UnicastSource::UnicastSource(Mode mode, IPEndPoint endpoint, const std::string& addr, unsigned short port)
: Source(mode),
  addr_(addr),
  port_(port),
  attached_(true)
{
  if(mode != INOUT)
  {
    throw UnsupportedModeException("The specified i/o mode is not valid for the current constructor");
  }

  create();
  bind(endpoint);
  state_ = ACTIVE;
}

UnicastSource::~UnicastSource()
{
  if(sockfd_ != INVALID_SOCKET)
  {
    close();
  }
}

void UnicastSource::create()
{
  sockfd_ = socket(AF_INET, SOCK_DGRAM, 0);
  if(sockfd_ == INVALID_SOCKET)
  {
    state_ = ERROR;
    throw OpenFailedException("Failed to create unicast source");
  }

  if(mode_ == IN || mode_ == INOUT)
  {
#ifdef WIN32
    readHandle_ = CreateEvent(NULL, FALSE, FALSE, NULL);
    if(WSAEventSelect(sockfd_, (WSAEVENT)readHandle_, FD_READ) == SOCKET_ERROR)
    {
      closesocket(sockfd_);
      sockfd_ = INVALID_SOCKET;
      state_ = ERROR;
      throw OpenFailedException("Failed to create read handle for unicast source");
    }
#else
    readHandle_ = sockfd_;
#endif
  }

  if(mode_ == OUT || mode_ == INOUT)
  {
#ifdef WIN32
    writeHandle_ = CreateEvent(NULL, FALSE, FALSE, NULL);
    if(WSAEventSelect(sockfd_, (WSAEVENT)writeHandle_, FD_WRITE) == SOCKET_ERROR)
    {
      closesocket(sockfd_);
      sockfd_ = INVALID_SOCKET;
      state_ = ERROR;
      throw OpenFailedException("Failed to create write handle for unicast source");
    }
#else
    writeHandle_ = sockfd_;
#endif
  }

  destination_.sin_family = AF_INET;
  if(attached_)
  {
    destination_.sin_port = htons(port_);
    if(addr_.empty())
    {
      destination_.sin_addr.s_addr = IN4ADDR_WILDCARD;
    }
    else
    {
      destination_.sin_addr.s_addr = inet_addr(addr_.c_str());
    }
  }
}

void UnicastSource::bind(const IPEndPoint& ie)
{
  SOCKADDR_IN sin;
  sin.sin_family = AF_INET;
  sin.sin_port = htons(ie.getPort());
  sin.sin_addr.s_addr = inet_addr(ie.getAddress().c_str());

  int result = ::bind(sockfd_, (const struct sockaddr*)&sin, sizeof(sin));

  if(result == SOCKET_ERROR)
  {
    state_ = ERROR;
    throw OpenFailedException("Failed to bind unicast source");
  }
}

void UnicastSource::close()
{
  closesocket(sockfd_);
  sockfd_ = INVALID_SOCKET;

#ifdef WIN32
  CloseHandle(readHandle_);
  CloseHandle(writeHandle_);
#endif

  readHandle_ = INVALID_HANDLE;
  writeHandle_ = INVALID_HANDLE;

  state_ = INACTIVE;
}

Format* UnicastSource::readRaw()
{
  SOCKADDR_IN sin;
  socklen_t sinlen = sizeof(sin);
  RawDatagramFormat* fmt = new RawDatagramFormat(RawFormat::DATAGRAM_BUFFER, this);

  int total = recvfrom(sockfd_, fmt->getBuffer(), fmt->getBufferSize(), 0, (struct sockaddr*)&sin, &sinlen);

  // Note that 0 is a valid datagram size
  if(total < 0)
  {
    delete fmt;
    state_ = ERROR;
    throw ReadFailedException("Error reading from unicast source");
  }

  fmt->setDataSize(total);
  fmt->setPort(ntohs(sin.sin_port));
  fmt->setAddress(inet_ntoa(sin.sin_addr));

  return fmt;
}

void UnicastSource::writeRaw(const Format* fmt)
{
  const RawFormat* rawfmt = NULL;
  if(attached_)
  {
    rawfmt = dynamic_cast< const RawFormat* >(fmt);
    if(rawfmt == NULL)
    {
      state_ = ERROR;
      throw UnsupportedFormatException("Unicast source could not process the specified format type for writing");
    }
  }
  else
  {
    const RawDatagramFormat* rdfmt = dynamic_cast< const RawDatagramFormat* >(fmt);
    if(rdfmt != NULL)
    {
      rawfmt = rdfmt;
      destination_.sin_port = htons(rdfmt->getPort());
      destination_.sin_addr.s_addr = inet_addr(rdfmt->getAddress().c_str());
    }
    else
    {
      state_ = ERROR;
      throw UnsupportedFormatException("Unicast source could not process the specified format type for writing; RawDatagramFormat required for unattached unicast source");
    }
  }

  // Note that 0 is a valid datagram size
  if(rawfmt->getDataSize() >= 0)
  {
    int total = sendto(sockfd_, rawfmt->getBuffer(), rawfmt->getDataSize(), 0, (const struct sockaddr*)&destination_, sizeof(destination_));

    if(total < 0)
    {
      state_ = ERROR;
      throw WriteFailedException("Error writing to unicast source");
    }
  }
}
