#ifndef BYTEORDER_H
#define BYTEORDER_H

#include <inttypes.h>
#include <algorithm>

namespace phylum {

// Determine byte order
#if !defined(CORE_BIG_ENDIAN) && !defined(CORE_LITTLE_ENDIAN)

  #if defined(_WIN32) || defined(_WIN64)
    // Check Windows endian type
    #include <rpcndr.h>
    #if NDR_LOCAL_ENDIAN == NDR_LITTLE_ENDIAN
      #define CORE_LITTLE_ENDIAN
    #endif
  #else
    // Check UNIX endian type
    #include <sys/param.h>
    #if __BYTE_ORDER == __LITTLE_ENDIAN
      #define CORE_LITTLE_ENDIAN
    #endif
  #endif

#endif

// Endian types
enum ByteOrder
{
  ORDER_LITTLE,
  ORDER_BIG
};

#ifdef CORE_LITTLE_ENDIAN
static const ByteOrder SYSTEMBYTEORDER = ORDER_LITTLE;
#else
static const ByteOrder SYSTEMBYTEORDER = ORDER_BIG;
#endif

// Byte swap functions
template <typename T>
inline T swap_bytes(T& value)
{
  T* ptr = &value;
  size_t size = sizeof(T) - 1;
  size_t half = sizeof(T) >> 1;
  for(size_t i = 0; i < half; i++)
  {
    std::swap(ptr[0], ptr[size - i]);
  }

  return value;
}

// Swap all elements of an array
template <typename T>
inline void swap_bytes_array(T* value, size_t length)
{
  for(size_t i = 0; i < length; i++)
  {
    swap_bytes(value[i]);
  }
}

// Network centric swaps
template <typename T>
inline T ntoh(T& value)
{
#ifdef CORE_LITTLE_ENDIAN
  return swap_bytes(value);
#else
  return value;
#endif
}

// Network centric swaps
template <typename T>
inline T hton(T& value)
{
#ifdef CORE_LITTLE_ENDIAN
  return swap_bytes(value);
#else
  return value;
#endif
}

// Specializations for the basic integer types
template <int8_t>
inline void swap_bytes(int8_t& value)
{
  return value;
}

template <uint8_t>
inline void swap_bytes(uint8_t& value)
{
  return value;
}

template <int16_t>
inline void swap_bytes(int16_t& value)
{
  value = (((value << 8) & 0x00FF) | ((value >> 8) & 0xFF00));
  return value;
}

template <uint16_t>
inline void swap_bytes(uint16_t& value)
{
  value = (((value << 8) & 0xFF00) | ((value >> 8) & 0x00FF));
  return value;
}

template <int32_t>
inline void swap_bytes(int32_t& value)
{
  value = (((value << 24) & 0xFF000000) | ((value << 16) & 0x00FF0000) |
           ((value >> 16) & 0x0000FF00) | ((value >> 24) & 0x000000FF));
  return value;
}

template <uint32_t>
inline void swap_bytes(uint32_t& value)
{
  value = (((value << 24) & 0xFF000000) | ((value << 16) & 0x00FF0000) |
           ((value >> 16) & 0x0000FF00) | ((value >> 24) & 0x000000FF));
  return value;
}

template <int8_t>
inline void swap_bytes_array(int8_t* value, size_t length)
{
  // NOOP
}

template <uint8_t>
inline void swap_bytes_array(uint8_t* value, size_t length)
{
  // NOOP
}

} // namespace phylum

#endif // BYTEORDER_H
