#ifdef WIN32
#include <windows.h>
#define TRETURN DWORD WINAPI
#else
#include <pthread.h>
#include <unistd.h>
#define TRETURN void*
#endif

#include <iostream>
#include <stdio.h>
#include <tbb/task_scheduler_init.h>
#include "Exceptions.h"
#include "ForwardTerminator.h"
#include "PipelineDispatcher.h"
#include "Pipeline.h"
#include "Monitor.h"
#include "UnicastSource.h"

// Run function for thread that executes the source monitor
TRETURN run_monitor(void* monitor)
{
  tbb::task_scheduler_init tsi;
  static_cast<Monitor*>(monitor)->run();
  return NULL;
}

// Run function for thread that executes the pipeline dispatcher
TRETURN run_dispatcher(void* dispatcher)
{
  tbb::task_scheduler_init tsi;
  static_cast<PipelineDispatcher*>(dispatcher)->run();
  return NULL;
}

// String tokenizer from http://www.oopweb.com/CPP/Documents/CPPHOWTO/Volume/C++Programming-HOWTO-7.html
void tokenize(const std::string& str, std::vector<std::string>& tokens, const std::string& delimiters = " ")
{
  // Skip delimiters at beginning.
  std::string::size_type lastPos = str.find_first_not_of(delimiters, 0);
  // Find first "non-delimiter".
  std::string::size_type pos = str.find_first_of(delimiters, lastPos);

  while (std::string::npos != pos || std::string::npos != lastPos)
  {
    // Found a token, add it to the vector.
    tokens.push_back(str.substr(lastPos, pos - lastPos));
    // Skip delimiters.  Note the "not_of"
    lastPos = str.find_first_not_of(delimiters, pos);
    // Find next "non-delimiter"
    pos = str.find_first_of(delimiters, lastPos);
  }
}

int main(int argc, char** argv)
{
  // Initialize task scheduler
  tbb::task_scheduler_init tsi;

  // Process arguments
  if(argc < 3)
  {
    std::cout << "usage: " << argv[0] << " [input-port[,input-port]*] [output-addr:output-port[,output-addr:output-port]*]" << std::endl;
    return -1;
  }

  typedef std::vector<std::string> Strings;
  Strings inputports;
  Strings outputports;

  tokenize(argv[1], inputports, ",");
  tokenize(argv[2], outputports, ",");

  // Define socket lists
  typedef std::vector<UnicastSource*> Sources;
  Sources inputs;
  Sources outputs;

  try
  {
    // Add the pipeline delegate for the source to the disptcher
    // This program is simple enough that each source may share a single pipeline,
    // as no source specific filtering is done.  When different input or output sources require
    // different filters for post/pre processing these sources will need separate pipelines
    Pipeline pipeline;
    PipelineDelegate delegate(&pipeline);

    // Create dispatcher and monitor
    PipelineDispatcher dispatch;
    Monitor monitor;

    // Create input sources
    for(Strings::const_iterator iter = inputports.begin(); iter != inputports.end(); ++iter)
    {
      UnicastSource* input = new UnicastSource(Source::IN, atoi((*iter).c_str()));
      inputs.push_back(input);

      //  Add the source and its delegate to the dispatcher
      dispatch.addSourceDelegate(input, &delegate);

      // Add the source and its delegate to the monitor
      monitor.addSourceMonitor(input, Monitor::READ, &delegate);
    }

    // Setup the pipeline with a terminator to forward to the output sources
    // The terminitor is setup as a non-serial stage of the pipeline,
    // beacause the datagrams to be forwarded are independent of each other
    // with respect to order of processing and delivery
    ForwardTerminator term(/*is_serial*/false);
    pipeline.setTerminator(&term);

    // Create output sources
    for(Strings::const_iterator iter = outputports.begin(); iter != outputports.end(); ++iter)
    {
      Strings endpoint;
      tokenize(*iter, endpoint, ":");
      UnicastSource* output = new UnicastSource(Source::OUT, endpoint[0], atoi(endpoint[1].c_str()));
      term.addSource(output);
    }

    // Start monitor and dispatcher threads
#ifdef WIN32
    HANDLE t1, t2;
    t1 = CreateThread(NULL, 0, run_monitor, (void*)&monitor, 0, NULL);
    t2 = CreateThread(NULL, 0, run_dispatcher, (void*)&dispatch, 0, NULL);
#else
    pthread_t t1, t2;
    pthread_create(&t1, NULL, run_monitor, (void*)&monitor);
    pthread_create(&t2, NULL, run_dispatcher, (void*)&dispatch);
#endif

    // Stop on keyboard input
    std::cout << "Forwarder is running.  Press Enter key to quit.  " << std::endl;

    getchar();

    // Stop the dispacther and the monitor
    dispatch.stop();
    monitor.stop();

#ifdef WIN32
    WaitForSingleObject(t1, INFINITE);
    WaitForSingleObject(t2, INFINITE);
#else
    pthread_join(t1, NULL);
    pthread_join(t2, NULL);
#endif

    // Deallocate the sources
    for(Sources::iterator iter = inputs.begin(); iter != inputs.end(); ++iter)
    {
      delete *iter;
    }

    for(Sources::iterator iter = outputs.begin(); iter != outputs.end(); ++iter)
    {
      delete *iter;
    }
  }
  catch(Exception e)
  {
    std::cout << e.getMessage() << std::endl;
  }

  return 0;
}
