#ifdef WIN32
#include <windows.h>
#define usleep(x) Sleep(x/1000)
#define TRETURN DWORD WINAPI
#else
#include <pthread.h>
#include <unistd.h>
#define TRETURN void*
#endif

#include <iostream>
#include <tbb/task_scheduler_init.h>
#include <tbb/spin_mutex.h>
#include "Format.h"
#include "Pipeline.h"
#include "PipelineFunctions.h"

using namespace std;

typedef tbb::spin_mutex IOMutex;
IOMutex iomutex;

struct DelayFormat : public Format
{
  int delay_;
  DelayFormat(int delay, Creator* creator = NULL) : Format(1, creator), delay_(delay) { }
  DelayFormat* clone(Creator*) const { return NULL; }
};

class Delay : public Filter
{
public:
  int id_;
  Delay(int id) : Filter(false), id_(id) { }
  virtual bool operator()(const Format* item)
  {
    IOMutex::scoped_lock lock;

    const DelayFormat* delay = dynamic_cast<const DelayFormat*>(item);

    lock.acquire(iomutex);
    cout << "Begin " << delay->delay_ << " microsecond delay #" << id_ << endl;
    lock.release();

    usleep(delay->delay_);

    lock.acquire(iomutex);
    cout << "End delay #" << id_ << endl;
    lock.release();

    return true;
  }
};

class PrintTerm : public Terminator
{
public:
  PrintTerm() : Terminator(true) { }

  virtual bool operator()(Format* item)
  {
    const DelayFormat* delay = dynamic_cast<const DelayFormat*>(item);
    IOMutex::scoped_lock lock(iomutex);
    cout << "Completed delay #" << delay->delay_ << endl;
    return true;
  };
};


// Test multithread modification of pipeline while running
// The pipeline can not be safely cleared while it is running,
// Clear must wait for
TRETURN run_pipeline(void* pipe)
{
  tbb::task_scheduler_init tsi;
  IOMutex::scoped_lock lock;

  lock.acquire(iomutex);
  cout << "Starting" << endl;
  lock.release();

  static_cast<Pipeline*>(pipe)->run(8);

  lock.acquire(iomutex);
  cout << "Done" << endl;
  lock.release();

  return NULL;
}

Delay ending(4);
DelayFormat d(200);

TRETURN clear_pipeline(void* pipe)
{
  tbb::task_scheduler_init tsi;
  IOMutex::scoped_lock lock;

  lock.acquire(iomutex);
  cout << "Clearing" << endl;
  lock.release();

  Pipeline* p = static_cast<Pipeline*>(pipe);
  p->suspend();
  p->clearFunctions();
  p->addFunction(&ending);
  p->addData(&d);

  lock.acquire(iomutex);
  cout << "Cleared" << endl;
  lock.release();

  usleep(10000000);

  p->resume();

  lock.acquire(iomutex);
  cout << "Resumed" << endl;
  lock.release();

  return NULL;
}

int main()
{
  tbb::task_scheduler_init tsi;

  // Check for parellelism in pipeline for a single item
  PrintTerm term;
  Delay one(1), two(2), three(3);
  DelayFormat a(1000), b(5000000), c(2000000);

  Pipeline pipe(&term);
  pipe.addFunction(&one);
  pipe.addFunction(&two);
  pipe.addFunction(&three);

  pipe.addData(&a);
  pipe.addData(&b);
  pipe.addData(&c);

#ifdef WIN32
  HANDLE t1, t2;
  t1 = CreateThread(NULL, 0, run_pipeline, (void*)&pipe, 0, NULL);
  t2 = CreateThread(NULL, 0, clear_pipeline, (void*)&pipe, 0, NULL);
  WaitForSingleObject(t1, INFINITE);
  WaitForSingleObject(t2, INFINITE);
#else
  pthread_t t1, t2;
  pthread_create(&t1, NULL, run_pipeline, (void*)&pipe);
  usleep(5000000);
  pthread_create(&t2, NULL, clear_pipeline, (void*)&pipe);

  pthread_join(t1, NULL);
  pthread_join(t2, NULL);
#endif

  pipe.addData(&a);
  pipe.addData(&b);
  pipe.addData(&c);
  pipe.run(4);

  cout << "Exiting program" << endl;

  return 0;
}
