#ifdef WIN32
#include <windows.h>
#define usleep(x) Sleep(x/1000)
#define TRETURN DWORD WINAPI
#else
#include <pthread.h>
#include <unistd.h>
#define TRETURN void*
#endif

#include <iostream>
#include <tbb/tick_count.h>
#include <tbb/task_scheduler_init.h>
#include "Exceptions.h"
#include "ForwardTerminator.h"
#include "PipelineDispatcher.h"
#include "Pipeline.h"
#include "Monitor.h"
#include "FileSource.h"
#include "TextLinesClassifier.h"
#include "TextLinesDecomposer.h"
#include "TextLineFormat.h"
#include "TextLineClassifier.h"
#include "DecomposingDelegateAdapter.h"
#include "BinaryAdapterFormat.h"
#include "BinaryAdapterClassifier.h"

using namespace std;
using namespace phylum;

TRETURN run_monitor(void* monitor)
{
  tbb::task_scheduler_init tsi;
  try
  {
    static_cast<Monitor*>(monitor)->run();
  }
  catch(Exception e)
  {
    std::cout << e.getMessage() << std::endl;
    exit(-1);
  }
  return NULL;
}

TRETURN run_dispatcher(void* dispatcher)
{
  tbb::task_scheduler_init tsi;
  try
  {
    static_cast<PipelineDispatcher*>(dispatcher)->run();
  }
  catch(Exception e)
  {
    std::cout << e.getMessage() << std::endl;
    exit(-1);
  }
  return NULL;
}

// Strip comment from a line
class StripCommentTransform : public Transform
{
public:
  // Constructor - set transform mode as non-serial
  StripCommentTransform() : Transform(false) { }

  // Filter based on format contents
  virtual bool operator()(Format* fmt)
  {
    TextLineFormat* line = dynamic_cast<TextLineFormat*>(fmt);
    if(line == NULL)
    {
      throw UnsupportedFormatException("Invalid format for StripCommentTransform; expected TextLineFormat");
    }

    std::string s = line->getLine();
    std::string::size_type pos = s.find_first_of('#');
    if(pos != std::string::npos)
    {
      line->setLine(s.substr(0, pos));
    }

    return true;
  }
};

// Transform to remove whitespace from begining and end
// Empty strings are filtered
class StripWhitespaceTransform : public Transform
{
public:
  // Constructor - set transform mode as non-serial
  StripWhitespaceTransform() : Transform(false) { }

  // Filter based on format contents
  virtual bool operator()(Format* fmt)
  {
    TextLineFormat* line = dynamic_cast<TextLineFormat*>(fmt);
    if(line == NULL)
    {
      throw UnsupportedFormatException("Invalid format for StripWhitespaceTransform; expected TextLineFormat");
    }

    std::string s = line->getLine();
    std::string::size_type first = s.find_first_not_of(" \t\n\r");

    if(first == std::string::npos)
    {
      return false;
    }

    line->setLine(s.substr(first, s.find_last_not_of(" \t\n\r")-first+1));

    return true;
  }
};

// Filter to remove lines with invalid characters
class ContentFilter : public Filter
{
protected:
  std::string validCharacters_;
public:
  // Constructor - set filter mode as non-serial
  ContentFilter(const std::string& validCharacters) : Filter(false), validCharacters_(validCharacters) { }

  // Filter based on format contents
  virtual bool operator()(const Format* fmt)
  {
    const TextLineFormat* line = dynamic_cast<const TextLineFormat*>(fmt);
    if(line == NULL)
    {
      throw UnsupportedFormatException("Invalid format for ContentFilter; expected TextLineFormat");
    }

    const std::string& s = line->getLine();
    return (s.find_first_not_of(validCharacters_) != std::string::npos) ? false : true;
  }
};

// Filter to check for the correct number of values per line
class CSVFilter : public Filter
{
protected:
  unsigned int num_;
public:
  // Constructor - set filter mode as non-serial
  CSVFilter(unsigned int num) : Filter(false), num_(num) { }

  // Filter based on format contents
  virtual bool operator()(const Format* fmt)
  {
    const TextLineFormat* line = dynamic_cast<const TextLineFormat*>(fmt);
    if(line == NULL)
    {
      throw UnsupportedFormatException("Invalid format for CSVFilter; expected TextLineFormat");
    }

    const std::string& s = line->getLine();
    unsigned int count = 0;
    for(std::string::const_iterator iter = s.begin(); iter != s.end(); ++iter)
    {
      if(*iter == ',')
      {
        if(++count > num_)
        {
          break;
        }
      }
    }

    return (num_ == count) ? true : false;
  }
};

// Transform to add a line feed to the end of a line
class AppendLineFeedTransform : public Transform
{
public:
  // Constructor - set transform mode as non-serial
  AppendLineFeedTransform() : Transform(false) { }

  // Filter based on format contents
  virtual bool operator()(Format* fmt)
  {
    TextLineFormat* line = dynamic_cast<TextLineFormat*>(fmt);
    if(line == NULL)
    {
      throw UnsupportedFormatException("Invalid format for AppendLineFeedTransform; expected TextLineFormat");
    }

    std::string s = line->getLine();
    s.push_back('\n');
    line->setLine(s);

    return true;
  }
};

// String tokenizer from http://www.oopweb.com/CPP/Documents/CPPHOWTO/Volume/C++Programming-HOWTO-7.html
void tokenize(const std::string& str, std::vector<std::string>& tokens, const std::string& delimiters = " ")
{
  // Skip delimiters at beginning.
  std::string::size_type lastPos = str.find_first_not_of(delimiters, 0);
  // Find first "non-delimiter".
  std::string::size_type pos = str.find_first_of(delimiters, lastPos);

  while (std::string::npos != pos || std::string::npos != lastPos)
  {
    // Found a token, add it to the vector.
    tokens.push_back(str.substr(lastPos, pos - lastPos));
    // Skip delimiters.  Note the "not_of"
    lastPos = str.find_first_not_of(delimiters, pos);
    // Find next "non-delimiter"
    pos = str.find_first_of(delimiters, lastPos);
  }
}

// Converts the CSV lines to XML data
// Uses BinaryAdapterFormat to store the XML data, simulating creation of binary data
class CSV2XMLTranslator : public Translator
{
protected:
  BinaryAdapterFormat template_;
public:
  CSV2XMLTranslator() : Translator(false), template_(0, false, this)
  {
    template_.addTermStringField("time", '\n');
    template_.addTermStringField("x", '\n');
    template_.addTermStringField("y", '\n');
    template_.addTermStringField("z", '\n');
    template_.addTermStringField("vx", '\n');
    template_.addTermStringField("vy", '\n');
    template_.addTermStringField("vz", '\n');
  }

  // Filter based on format contents
  virtual Format* operator()(const Format* fmt)
  {
    const TextLineFormat* line = dynamic_cast<const TextLineFormat*>(fmt);
    if(line == NULL)
    {
      throw UnsupportedFormatException("Invalid format for CSV2XMLTranslator; expected TextLineFormat");
    }

    // Separate the CSV line to a vectr of string values
    std::vector<std::string> tokens;
    std::string s;
    tokenize(line->getLine(), tokens, ",");

    BinaryAdapterFormat* adapter = template_.clone(this);

    // Mix XML tags with data
    s = std::string("<Data>\n\t<Time>") + tokens[0] + std::string("</Time>");
    adapter->setFieldValue("time", s);

    s = std::string("\t<PositionX>") + tokens[1] + std::string("</PositionX>");
    adapter->setFieldValue("x", s);

    s = std::string("\t<PositionY>") + tokens[2] + std::string("</PositionY>");
    adapter->setFieldValue("y", s);

    s = std::string("\t<PositionZ>") + tokens[3] + std::string("</PositionZ>");
    adapter->setFieldValue("z", s);

    s = std::string("\t<VelocityX>") + tokens[4] + std::string("</VelocityX>");
    adapter->setFieldValue("vx", s);

    s = std::string("\t<VelocityY>") + tokens[5] + std::string("</VelocityY>");
    adapter->setFieldValue("vy", s);

    s = std::string("\t<VelocityZ>") + tokens[6] + std::string("</VelocityZ>\n</Data>");
    adapter->setFieldValue("vz", s);

    return adapter;
  }
};

int main()
{
  tbb::task_scheduler_init tsi;

  tbb::tick_count start = tbb::tick_count::now();

  try
  {
    // Create input and output sources
    FileSource csvin(FileSource::IN, "input.csv");
    FileSource txtin(FileSource::IN, "input.txt");

    FileSource lineout1(FileSource::OUT, "lineoutput1.csv", FileSource::CREAT|FileSource::TRUNC);
    FileSource lineout2(FileSource::OUT, "lineoutput2.csv", FileSource::CREAT|FileSource::TRUNC);

    FileSource xmlout1(FileSource::OUT, "xmloutput1.txt", FileSource::CREAT|FileSource::TRUNC);
    FileSource xmlout2(FileSource::OUT, "xmloutput2.txt", FileSource::CREAT|FileSource::TRUNC);

    // Create filters
    StripCommentTransform linesct, xmlsct;                                  // Remove all comments begining with '#'
    StripWhitespaceTransform lineswt, xmlswt;                               // Remove all whitespace from begining and end of line; lines only containing white space are eliminated from pipeline
    ContentFilter linecflt("0123456789,.+-e"), xmlcflt("0123456789,.+-e");  // Reject lines containing invalid values
    CSVFilter linecsvflt(6), xmlcsvflt(6);                                  // Reject lines without exactly 7 columns (6 commas)
    FilterGroup linefgrp(false), xmlfgrp(false);                            // Process the filters at the same time
    linefgrp.addFilter(&linecflt);
    linefgrp.addFilter(&linecsvflt);
    xmlfgrp.addFilter(&xmlcflt);
    xmlfgrp.addFilter(&xmlcsvflt);
    AppendLineFeedTransform alft;          // Append '\n' to the line before writing to a file
    CSV2XMLTranslator xmlt;
    ForwardTerminator lineterm(true);      // Write the reults to one or more files
    ForwardTerminator xmlterm(true);       // Write the reults to one or more files

    // Create pipelines
    Pipeline linepipeline;
    Pipeline xmlpipeline;

    // Add filters to text-line based pipeline and the xml pipeline
    linepipeline.addFunction(&linesct);
    linepipeline.addFunction(&lineswt);
    linepipeline.addFunction(&linefgrp);
    linepipeline.addFunction(&alft);

    xmlpipeline.addFunction(&xmlsct);
    xmlpipeline.addFunction(&xmlswt);
    xmlpipeline.addFunction(&xmlfgrp);
    xmlpipeline.addFunction(&xmlt);

    // Setup the pipeline with a terminator to forward to the output source
    lineterm.addSource(&lineout1);
    lineterm.addSource(&lineout2);
    linepipeline.setTerminator(&lineterm);

    xmlterm.addSource(&xmlout1);
    xmlterm.addSource(&xmlout2);
    xmlpipeline.setTerminator(&xmlterm);

    // Add pipeline to delegate
    Pipelines pipelines;
    pipelines.push_back(&linepipeline);
    pipelines.push_back(&xmlpipeline);
    PipelinesDelegate delegate(&pipelines);

    TextLinesDecomposer tld;
    DecomposingDelegateAdapter dda(&delegate, &tld);

    // Add the pipeline delegate for the source to the disptcher
    PipelineDispatcher dispatch;
    dispatch.addSourceDelegate(&csvin, &dda);

    // Set the classifier for the source
    TextLinesClassifier tlsc;
    csvin.setClassifier(&tlsc);

    TextLineClassifier tlc;
    lineout1.setClassifier(&tlc);
    lineout2.setClassifier(&tlc);

    BinaryAdapterClassifier bac;
    xmlout1.setClassifier(&bac);
    xmlout2.setClassifier(&bac);

    // Add the source and its delegate to the monitor
    Monitor monitor;
    monitor.addSourceMonitor(&csvin, Monitor::READ, &dda);
    monitor.addSourceMonitor(&lineout1, Monitor::WRITE, NULL);
    monitor.addSourceMonitor(&lineout2, Monitor::WRITE, NULL);
    monitor.addSourceMonitor(&xmlout1, Monitor::WRITE, NULL);
    monitor.addSourceMonitor(&xmlout2, Monitor::WRITE, NULL);

    // Start monitor and dispatcher threads
#ifdef WIN32
    HANDLE t1, t2;
    t1 = CreateThread(NULL, 0, run_monitor, (void*)&monitor, 0, NULL);
    t2 = CreateThread(NULL, 0, run_dispatcher, (void*)&dispatch, 0, NULL);
#else
    pthread_t t1, t2;
    pthread_create(&t1, NULL, run_monitor, (void*)&monitor);
    pthread_create(&t2, NULL, run_dispatcher, (void*)&dispatch);
#endif

    // Wait for eof
    while(csvin.getState() != Source::DONE)
    {
      usleep(100);
    }

    // Wait for pipeline data que to clear
    while(linepipeline.getDataSize() > 0 || xmlpipeline.getDataSize() > 0)
    {
      usleep(100);
    }

    // Wait for write queues to clear
    while(lineout1.getWriteQueueSize() > 0 || lineout2.getWriteQueueSize() > 0 || xmlout1.getWriteQueueSize() > 0 || xmlout2.getWriteQueueSize() > 0)
    {
      usleep(100);
    }

    monitor.stop();
    dispatch.stop();

#ifdef WIN32
    WaitForSingleObject(t1, INFINITE);
    WaitForSingleObject(t2, INFINITE);
#else
    pthread_join(t1, NULL);
    pthread_join(t2, NULL);
#endif
  }
  catch(Exception e)
  {
    std::cout << e.getMessage() << std::endl;
  }

  std::cout << "File processed in " << (tbb::tick_count::now() - start).seconds() << " seconds" << std::endl;

  return 0;
}
